"""Module of objects that resemble or contain a profile, i.e. a 1 or 2-D f(x) representation."""
from __future__ import annotations

import enum
import math
import warnings
from typing import Union, Tuple, Sequence, List, Optional

import argue
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Circle as mpl_Circle
from scipy import ndimage, signal
from scipy.interpolate import interp1d
from scipy.ndimage import gaussian_filter1d
from scipy.optimize import OptimizeWarning, minimize
from scipy.stats import linregress

from .geometry import Point, Circle
from .hill import Hill
from .utilities import convert_to_enum

# for Hill fits of 2D device data the # of points can be small.
# This results in optimization warnings about the variance of the fit (the variance isn't of concern for us for that particular item)
warnings.simplefilter("ignore", OptimizeWarning)


def gamma_1d(
    reference: np.ndarray,
    evaluation: np.ndarray,
    dose_to_agreement: float = 1,
    distance_to_agreement: int = 1,
    gamma_cap_value: float = 2,
    global_dose: bool = True,
    dose_threshold: float = 5,
    fill_value: float = np.nan,
) -> np.ndarray:
    """Perform a 1D gamma of two 1D profiles/arrays. This does NOT check lengths or
    spatial consistency. It performs an element-by-element evaluation. It is the responsibility
    of the caller to ensure the reference and evaluation have comparable spatial resolution.

    The algorithm follows Table I of D. Low's 2004 paper: Evaluation of the gamma dose distribution comparison method: https://aapm.onlinelibrary.wiley.com/doi/epdf/10.1118/1.1598711

    Parameters
    ----------

    reference
        The reference profile.
    evaluation
        The evaluation profile.
    dose_to_agreement
        The dose to agreement in %. E.g. 1 is 1% of global reference max dose.
    distance_to_agreement
        The distance to agreement in **elements**. E.g. if the value is 4 this means 4 elements from the reference point under calculation.
        Must be >0
    gamma_cap_value
        The value to cap the gamma at. E.g. a gamma of 5.3 will get capped to 2. Useful for displaying data with a consistent range.
    global_dose
        Whether to evaluate the dose to agreement threshold based on the global max or the dose point under evaluation.
    dose_threshold
        The dose threshold as a number between 0 and 100 of the % of max dose under which a gamma is not calculated.
        This is not affected by the global/local dose normalization and the threshold value is evaluated against the global max dose, period.
    fill_value
        The value to give pixels that were not calculated because they were under the dose threshold. Default
        is NaN, but another option would be 0. If NaN, allows the user to calculate mean/median gamma over just the
        evaluated portion and not be skewed by 0's that should not be considered.
    """
    if reference.ndim != 1 or evaluation.ndim != 1:
        raise ValueError(
            f"Reference and evaluation arrays must be 1D. Got reference: {reference.ndim} and evaluation: {evaluation.ndim}"
        )
    threshold = reference.max() / 100 * dose_threshold
    # convert dose to agreement to % of global max; ignored later if local dose
    dose_ta = dose_to_agreement / 100 * reference.max()
    # pad eval array on both edges so our search does not go out of bounds
    eval_padded = np.pad(evaluation, distance_to_agreement, mode="edge")
    # iterate over each reference element, computing distance value and dose value
    gamma = []
    for r_idx, ref_point in enumerate(reference):
        # skip if below dose threshold
        if ref_point < threshold:
            gamma.append(fill_value)
            continue
        # we search at the same indices in eval_padded, but remember eval_padded has extra indices on each edge,
        # so this is actually searching from -DTA to +DTA because r_idx in eval_padded is off by distance_to_agreement.
        capital_gammas = []
        for e_idx, eval_point in enumerate(
            eval_padded[r_idx : r_idx + 2 * distance_to_agreement + 1]
        ):
            dist = abs(e_idx - distance_to_agreement)
            dose = eval_point - ref_point
            if not global_dose:
                dose_ta = dose_to_agreement / 100 * ref_point
            capital_gamma = math.sqrt(
                dist ** 2 / distance_to_agreement ** 2 + dose ** 2 / dose_ta ** 2
            )
            capital_gammas.append(capital_gamma)
        gamma.append(min(min(capital_gammas), gamma_cap_value))
    return np.asarray(gamma)


def stretch(
    array: np.ndarray, min: int = 0, max: int = 1, fill_dtype: Optional[np.dtype] = None
) -> np.ndarray:
    """'Stretch' the profile to the fit a new min and max value and interpolate in between.
    From: http://www.labri.fr/perso/nrougier/teaching/numpy.100/  exercise #17

    Parameters
    ----------
    array: numpy.ndarray
        The numpy array to stretch.
    min : number
        The new minimum of the values.
    max : number
        The new maximum value.
    fill_dtype : numpy data type
        If None (default), the array will be stretched to the passed min and max.
        If a numpy data type (e.g. np.int16), the array will be stretched to fit the full range of values
        of that data type. If a value is given for this parameter, it overrides ``min`` and ``max``.
    """
    new_max = max
    new_min = min
    if fill_dtype is not None:
        try:
            di = np.iinfo(fill_dtype)
        except ValueError:
            di = np.finfo(fill_dtype)
        new_max = di.max
        new_min = di.min
    # perfectly normalize the array (0..1). ground, then div by range
    stretched_array = (array - array.min()) / (array.max() - array.min())
    # stretch normalized array to new max/min
    stretched_array *= new_max
    # stretched_array += new_min
    if fill_dtype:
        stretched_array = stretched_array.astype(fill_dtype)
    return stretched_array


class ProfileMixin:
    """A mixin to provide various manipulations of 1D profile data."""

    values: np.ndarray

    def invert(self) -> None:
        """Invert (imcomplement) the profile."""
        orig_array = self.values
        self.values = -orig_array + orig_array.max() + orig_array.min()

    def normalize(self, norm_val: Union[str, float] = "max") -> None:
        """Normalize the profile to the given value.

        Parameters
        ----------
        norm_val : str, number
            If a string, must be 'max', which normalizes the values to the maximum value.
            If a number, normalizes all values to that number.
        """
        if norm_val == "max":
            val = self.values.max()
        else:
            val = norm_val
        self.values /= val

    def stretch(self, min: float = 0, max: float = 1) -> None:
        """'Stretch' the profile to the min and max parameter values.

        Parameters
        ----------
        min : number
            The new minimum of the values
        max : number
            The new maximum value.
        """
        self.values = stretch(self.values, min=min, max=max)

    def ground(self) -> float:
        """Ground the profile such that the lowest value is 0.

        Returns
        -------
        float
            The minimum value that was used as the grounding value.
        """
        min_val = self.values.min()
        self.values = self.values - min_val
        return min_val

    @argue.options(kind=("median", "gaussian"))
    def filter(self, size: float = 0.05, kind: str = "median") -> None:
        """Filter the profile.

        Parameters
        ----------
        size : float, int
            Size of the median filter to apply.
            If a float, the size is the ratio of the length. Must be in the range 0-1.
            E.g. if size=0.1 for a 1000-element array, the filter will be 100 elements.
            If an int, the filter is the size passed.
        kind : {'median', 'gaussian'}
            The kind of filter to apply. If gaussian, `size` is the sigma value.
        """
        if isinstance(size, float):
            if 0 < size < 1:
                size = int(round(len(self.values) * size))
                size = max(size, 1)
            else:
                raise TypeError("Float was passed but was not between 0 and 1")

        if kind == "median":
            self.values = ndimage.median_filter(self.values, size=size)
        elif kind == "gaussian":
            self.values = ndimage.gaussian_filter(self.values, sigma=size)

    def __len__(self):
        return len(self.values)

    def __getitem__(self, items):
        return self.values[items]


class Interpolation(enum.Enum):
    """Interpolation Enum"""

    NONE = None  #:
    LINEAR = "Linear"  #:
    SPLINE = "Spline"  #:


class Normalization(enum.Enum):
    """Normalization method Enum"""

    NONE = None  #:
    GEOMETRIC_CENTER = "Geometric center"  #:
    BEAM_CENTER = "Beam center"  #:
    MAX = "Max"  #:


class Edge(enum.Enum):
    """Edge detection Enum"""

    FWHM = "FWHM"  #:
    INFLECTION_DERIVATIVE = "Inflection Derivative"  #:
    INFLECTION_HILL = "Inflection Hill"  #:


class SingleProfile(ProfileMixin):
    """A profile that has one large signal, e.g. a radiation beam profile.
    Signal analysis methods are given, mostly based on FWXM and on Hill function calculations.
    Profiles with multiple peaks are better suited by the MultiProfile class.
    """

    def __init__(
        self,
        values: np.ndarray,
        dpmm: float = None,
        interpolation: Union[Interpolation, str, None] = Interpolation.LINEAR,
        ground: bool = True,
        interpolation_resolution_mm: float = 0.1,
        interpolation_factor: float = 10,
        normalization_method: Union[Normalization, str] = Normalization.BEAM_CENTER,
        edge_detection_method: Union[Edge, str] = Edge.FWHM,
        edge_smoothing_ratio: float = 0.003,
        hill_window_ratio: float = 0.1,
        x_values: Optional[np.ndarray] = None,
    ):
        """
        Parameters
        ----------
        values
            The profile numpy array. Must be 1D.
        dpmm
            The dots (pixels) per mm. Pass to get info like beam width in distance units in addition to pixels
        interpolation
            Interpolation technique.
        ground
            Whether to ground the profile (set min value to 0). Helpful most of the time.
        interpolation_resolution_mm
            The resolution that the interpolation will scale to. **Only used if dpmm is passed and interpolation is set**.
            E.g. if the dpmm is 0.5 and the resolution is set to 0.1mm the data will be interpolated to have a new dpmm of 10 (1/0.1).
        interpolation_factor
            The factor to multiply the data by. **Only used if interpolation is used and dpmm is NOT passed**. E.g. 10
            will perfectly decimate the existing data according to the interpolation method passed.
        normalization_method
            How to pick the point to normalize the data to.
        edge_detection_method
            The method by which to detect the field edge. FWHM is reasonable most of the time except for FFF beams.
            Inflection-derivative will use the max gradient to determine the field edge. Note that this may not be the
            50% height. In fact, for FFF beams it shouldn't be. Inflection methods are better for FFF and other unusual
            beam shapes.
        edge_smoothing_ratio
            **Only applies to INFLECTION_DERIVATIVE and INFLECTION_HILL.**

            The ratio of the length of the values to use as the sigma for a Gaussian filter applied before searching for
            the inflection. E.g. 0.005 with a profile of 1000 points will result in a sigma of 5.
            This helps make the inflection point detection more robust to noise. Increase for noisy data.
        hill_window_ratio
            The ratio of the field size to use as the window to fit the Hill function. E.g. 0.2 will using a window
            centered about each edge with a width of 20% the size of the field width. **Only applies when the edge
            detection is INFLECTION_HILL**.
        x_values
            The x-values of the profile, if any. If None, will generate a simple range(len(values)).
        """
        self._interp_method = convert_to_enum(interpolation, Interpolation)
        self._interpolation_res = interpolation_resolution_mm
        self._interpolation_factor = interpolation_factor
        self._norm_method = convert_to_enum(normalization_method, Normalization)
        self._edge_method = convert_to_enum(edge_detection_method, Edge)
        self._edge_smoothing_ratio = edge_smoothing_ratio
        self._hill_window_ratio = hill_window_ratio
        self.values = (
            values  # set initial data so we can do things like find beam center
        )
        self.dpmm = dpmm
        fitted_values, new_dpmm, x_indices = self._interpolate(
            values,
            x_values,
            dpmm,
            interpolation_resolution_mm,
            interpolation_factor,
            self._interp_method,
        )
        self.values = fitted_values
        self.x_indices = x_indices
        self._x_interp1d = interp1d(list(range(len(x_indices))), x_indices)
        self._ground = ground
        if ground:
            fitted_values -= fitted_values.min()
        self._y_interp1d = interp1d(
            x_indices, fitted_values, bounds_error=False, fill_value="extrapolate"
        )
        norm_values = self._normalize(fitted_values, self._norm_method)
        self.values = norm_values  # update values
        self._y_interp1d = interp1d(
            x_indices, norm_values, bounds_error=False, fill_value="extrapolate"
        )

    def _x_interp_to_original(
        self, location: Union[float, np.ndarray]
    ) -> Union[float, np.ndarray]:
        """Get the x-value of the (possibly) interpolated profile. The input value is in the original
        value range. E.g. a profile with x-range of 0-10 is interpolated to 10x. Asking for the location at 99 would scale back to 9.9.
        We need this function because peak finding is independent of the x-values. I.e. peaks are found and reported according
        to the (0, len(x_values)) range. If the x-values are interpolated we need to get back to the original x-value."""
        x = self._x_interp1d(location)
        if isinstance(location, (float, int)) or location.size == 1:
            return float(x)
        return x

    def _y_original_to_interp(
        self, location: Union[float, np.ndarray]
    ) -> Union[float, np.ndarray]:
        """Get the interpolated y-value of the profile. This is a corollary to the _x_interp... function"""
        y = self._y_interp1d(location)
        if isinstance(location, (float, int)) or location.size == 1:
            return float(y)
        return y

    def resample(
        self, interpolation_factor: int = 10, interpolation_resolution_mm: float = 0.1
    ) -> SingleProfile:
        """Resample the profile at a new resolution. Returns a new profile"""
        # we have to set the dpmm to what it currently is after interpolating to resample.
        if self.dpmm:
            dpmm = 1 / self._interpolation_res
        else:
            dpmm = None
        return SingleProfile(
            values=self.values,
            x_values=self.x_indices,
            dpmm=dpmm,
            interpolation=self._interp_method,
            ground=self._ground,
            interpolation_resolution_mm=interpolation_resolution_mm,
            interpolation_factor=interpolation_factor,
            normalization_method=self._norm_method,
            edge_detection_method=self._edge_method,
            edge_smoothing_ratio=self._edge_smoothing_ratio,
            hill_window_ratio=self._hill_window_ratio,
        )

    @staticmethod
    def _interpolate(
        values,
        x_values,
        dpmm,
        interpolation_resolution,
        interpolation_factor,
        interp_method: Interpolation,
    ) -> (np.ndarray, float, float, float):
        """Fit the data to the passed interpolation method. Will also calculate the new values to correct the measurements such as dpmm"""
        if x_values is None:
            x_values = np.array(range(len(values)))
        if interp_method == Interpolation.NONE:
            return values, dpmm, x_values  # do nothing
        else:
            if dpmm is not None:
                samples = int(round(len(x_values) / (dpmm * interpolation_resolution)))
                new_dpmm = 1 / interpolation_resolution
            else:
                samples = int(round(len(x_values) * interpolation_factor))
                new_dpmm = None
            # Warning: BMF ahead
            # the problem is that if we upsample, the left and right ends are not equally sampled.
            # E.g. upsampling a 3-pixel array (0, 1, 2) by 10 normally results in ~20 elements. You
            # interpolate between 0 and 1, and 1 and 2.
            # The first issue is that you do not have a simple X proportion of
            # elements (3 * 10 = 30 but we get 20). Additionally, if these are pixels they have a
            # finite, physical size and technically those values are at the center of the pixels.
            # Thus, you actually need to sample beyond the left and right edges. In the
            # above case you'd really need to sample from approximately -0.5 to 2.5 to get ~10 pixels
            # for each original pixel. We also need to offset the x-values to be back to 0 again from -0.5.
            # We solve this by offsetting the new x-values by a proportion of the sampling ratio.
            # A ratio of 1 (identical sampling) should not have any offset and return the same values
            # As the ratio goes up, we approach the limit of 0.5 pixels. This follows a proportional relationship
            # with the ratio.
            resampling_factor = samples / len(values)
            offset = 0.5 - 1 / (2 * resampling_factor)
            if interp_method == Interpolation.LINEAR:
                kind = "linear"
            elif interp_method == Interpolation.SPLINE:
                kind = "cubic"
            f = interp1d(
                x_values,
                values,
                kind=kind,
                bounds_error=False,
                fill_value="extrapolate",
            )
            new_x = np.linspace(
                x_values[0] - offset, x_values[-1] + offset, num=samples
            )
            new_y = f(new_x)
            return new_y, new_dpmm, new_x

    def _normalize(self, values, method: Normalization) -> np.ndarray:
        """Normalize the data given a method."""
        if method == Normalization.NONE:
            return values
        elif method == Normalization.MAX:
            return values / values.max()
        elif method == Normalization.GEOMETRIC_CENTER:
            return values / self._geometric_center(values)["value (exact)"]
        elif method == Normalization.BEAM_CENTER:
            return values / self.beam_center()["value (@rounded)"]

    def _geometric_center(self, values) -> dict:
        """Returns the center index and value of the profile.

        If the profile has an even number of values the centre lies between the two centre indices and the centre
        value is the average of the two centre values else the centre index and value are returned."""
        plen = values.shape[0]
        # buffer overflow can cause the below addition to give strange results
        values = values.astype(np.float64)
        if plen % 2 == 0:  # plen is even and central detectors straddle CAX
            cax = (values[int(plen / 2)] + values[int(plen / 2) - 1]) / 2.0
        else:  # plen is odd and we have a central detector
            cax = values[int((plen - 1) / 2)]
        plen = (plen - 1) / 2.0
        return {"index (exact)": self._x_interp_to_original(plen), "value (exact)": cax}

    def geometric_center(self) -> dict:
        """The geometric center (i.e. the device center)"""
        return self._geometric_center(self.values)

    def beam_center(self) -> dict:
        """The center of the detected beam. This can account for asymmetries in the beam position (e.g. offset jaws)"""
        if self._edge_method == Edge.FWHM:
            data = self.fwxm_data(x=50)
            return {
                "index (rounded)": data["center index (rounded)"],
                "index (exact)": data["center index (exact)"],
                "value (@rounded)": data["center value (@rounded)"],
            }
        elif self._edge_method in (Edge.INFLECTION_DERIVATIVE, Edge.INFLECTION_HILL):
            infl = self.inflection_data()
            mid_point = (
                infl["left index (exact)"]
                + (infl["right index (exact)"] - infl["left index (exact)"]) / 2
            )
            return {
                "index (rounded)": int(round(mid_point)),
                "index (exact)": mid_point,
                "value (@rounded)": self._y_original_to_interp(int(round(mid_point))),
            }

    @argue.bounds(x=(0, 100))
    def fwxm_data(self, x: int = 50) -> dict:
        """Return the width at X-Max, where X is the percentage height.

        Parameters
        ----------
        x
            The percent height of the profile. E.g. x = 50 is 50% height,
            i.e. FWHM.
        """
        _, peak_props = find_peaks(self.values, fwxm_height=x / 100, max_number=1)
        left_idx = float(self._x_interp_to_original(peak_props["left_ips"][0]))
        right_idx = float(self._x_interp_to_original(peak_props["right_ips"][0]))
        width = right_idx - left_idx
        fwxm_center_idx = (right_idx - left_idx) / 2 + left_idx

        data = {
            "width (exact)": width,
            "width (rounded)": int(round(width)),
            "center index (rounded)": int(round(fwxm_center_idx)),
            "center index (exact)": fwxm_center_idx,
            "center value (@rounded)": float(
                self._y_original_to_interp(int(round(fwxm_center_idx)))
            ),
            "left index (exact)": left_idx,
            "left index (rounded)": int(round(left_idx)),
            "left value (@rounded)": float(
                self._y_original_to_interp(int(round(left_idx)))
            ),
            "right index (exact)": right_idx,
            "right index (rounded)": int(round(right_idx)),
            "right value (@rounded)": float(
                self._y_original_to_interp(int(round(right_idx)))
            ),
            "field values": self._y_original_to_interp(
                self.x_indices[int(round(left_idx)) : int(round(right_idx))]
            ),
            "peak_props": peak_props,
        }
        if self.dpmm:
            data["width (exact) mm"] = data["width (exact)"] / self.dpmm
            data["left distance (exact) mm"] = (
                abs(data["center index (exact)"] - data["left index (exact)"])
                / self.dpmm
            )
            data["right distance (exact) mm"] = (
                abs(data["right index (exact)"] - data["center index (exact)"])
                / self.dpmm
            )

        return data

    @argue.bounds(in_field_ratio=(0, 1.0), slope_exclusion_ratio=(0, 1.0))
    def field_data(
        self, in_field_ratio: float = 0.8, slope_exclusion_ratio=0.2
    ) -> dict:
        """Return the width at X-Max, where X is the percentage height.

        Parameters
        ----------
        in_field_ratio
            In Field Ratio: 1.0 is the entire detected field; 0.8 would be the central 80%, etc.
        slope_exclusion_ratio
            Ratio of the field width to use as the cutoff between "top" calculation and "slope" calculation. Useful for FFF beams.
            This area is centrally located in the field. E.g. 0.2 will use the central 20% of the field to calculate
            the "top" value. To calculate the slope of each side, the field width between the edges of the in_field_ratio
            and the slope exclusion region are used.

            .. warning:: The "top" value is always calculated. For FFF beams this should be reasonable, but for flat beams
                         this value may end up being non-sensible.
        """
        if slope_exclusion_ratio >= in_field_ratio:
            raise ValueError(
                "The exclusion region must be smaller than the field ratio"
            )
        if self._edge_method == Edge.FWHM:
            data = self.fwxm_data(x=50)
            beam_center_idx = data["center index (exact)"]
            full_width = data["width (exact)"]

        elif self._edge_method in (Edge.INFLECTION_DERIVATIVE, Edge.INFLECTION_HILL):
            infl_data = self.inflection_data()
            beam_center_idx = self.beam_center()["index (exact)"]
            full_width = (
                infl_data["right index (exact)"] - infl_data["left index (exact)"]
            )
        beam_center_idx_r = int(round(beam_center_idx))

        cax_idx = self.geometric_center()["index (exact)"]
        cax_idx_r = int(round(cax_idx))

        field_left_idx = beam_center_idx - in_field_ratio * full_width / 2
        field_left_idx_r = int(round(field_left_idx))
        field_right_idx = beam_center_idx + in_field_ratio * full_width / 2
        field_right_idx_r = int(round(field_right_idx))
        field_width = field_right_idx - field_left_idx

        # slope calcs
        inner_left_idx = beam_center_idx - slope_exclusion_ratio * field_width / 2
        inner_left_idx_r = int(round(inner_left_idx))
        inner_right_idx = beam_center_idx + slope_exclusion_ratio * field_width / 2
        inner_right_idx_r = int(round(inner_right_idx))
        left_fit = linregress(
            range(field_left_idx_r, inner_left_idx_r),
            self._y_original_to_interp(np.arange(field_left_idx_r, inner_left_idx_r)),
        )
        right_fit = linregress(
            range(inner_right_idx_r, field_right_idx_r),
            self._y_original_to_interp(np.arange(inner_right_idx_r, field_right_idx_r)),
        )

        # top calc
        fit_params = np.polyfit(
            range(inner_left_idx_r, inner_right_idx_r),
            self._y_original_to_interp(np.arange(inner_left_idx_r, inner_right_idx_r)),
            deg=2,
        )
        width = abs(inner_right_idx_r - inner_left_idx_r)

        def poly_func(x):
            # return the negative since we're MINIMIZING and want the top value
            return -(fit_params[0] * (x ** 2) + fit_params[1] * x + fit_params[2])

        # minimize the polynomial function
        min_f = minimize(
            poly_func,
            x0=(inner_left_idx_r + width / 2,),
            bounds=((inner_left_idx_r, inner_right_idx_r),),
        )
        top_idx = min_f.x[0]
        top_val = -min_f.fun

        data = {
            "width (exact)": field_width,
            "beam center index (exact)": beam_center_idx,
            "beam center index (rounded)": beam_center_idx_r,
            "beam center value (@rounded)": self._y_original_to_interp(
                round(beam_center_idx)
            ),
            "cax index (exact)": cax_idx,
            "cax index (rounded)": cax_idx_r,
            "cax value (@rounded)": self._y_original_to_interp(round(cax_idx)),
            "left index (exact)": field_left_idx,
            "left index (rounded)": field_left_idx_r,
            "left value (@rounded)": self._y_original_to_interp(round(field_left_idx)),
            "left slope": left_fit.slope,
            "left intercept": left_fit.intercept,
            "right slope": right_fit.slope,
            "right intercept": right_fit.intercept,
            "left inner index (exact)": inner_left_idx,
            "left inner index (rounded)": inner_left_idx_r,
            "right inner index (exact)": inner_right_idx,
            "right inner index (rounded)": inner_right_idx_r,
            '"top" index (exact)': top_idx,
            '"top" index (rounded)': int(round(top_idx)),
            '"top" value (@exact)': top_val,
            "top params": fit_params,
            "right index (exact)": field_right_idx,
            "right index (rounded)": field_right_idx_r,
            "right value (@rounded)": self._y_original_to_interp(
                round(field_right_idx)
            ),
            "field values": self._y_original_to_interp(
                np.arange(int(round(field_left_idx)), int(round(field_right_idx)))
            ),
        }
        if self.dpmm:
            data["width (exact) mm"] = data["width (exact)"] / self.dpmm
            data["left slope (%/mm)"] = data["left slope"] * self.dpmm * 100
            data["right slope (%/mm)"] = data["right slope"] * self.dpmm * 100
            data["left distance->beam center (exact) mm"] = (
                abs(data["beam center index (exact)"] - data["left index (exact)"])
                / self.dpmm
            )
            data["right distance->beam center (exact) mm"] = (
                abs(data["right index (exact)"] - data["beam center index (exact)"])
                / self.dpmm
            )
            data["left distance->CAX (exact) mm"] = (
                abs(data["cax index (exact)"] - data["left index (exact)"]) / self.dpmm
            )
            data["right distance->CAX (exact) mm"] = (
                abs(data["cax index (exact)"] - data["right index (exact)"]) / self.dpmm
            )

            data["left distance->top (exact) mm"] = (
                abs(data['"top" index (exact)'] - data["left index (exact)"])
                / self.dpmm
            )
            data["right distance->top (exact) mm"] = (
                abs(data['"top" index (exact)'] - data["right index (exact)"])
                / self.dpmm
            )
            data['"top"->beam center (exact) mm'] = (
                data['"top" index (exact)'] - data["beam center index (exact)"]
            ) / self.dpmm
            data['"top"->CAX (exact) mm'] = (
                abs(data['"top" index (exact)'] - data["cax index (exact)"]) / self.dpmm
            )
        return data

    def inflection_data(self) -> dict:
        """Calculate the profile inflection values using either the 2nd derivative or a fitted Hill function.

        .. note::
            This only applies if the edge detection method is `INFLECTION_...`.

        Parameters
        ----------

        """
        # get max/min of the gradient, which is basically the same as the 2nd deriv 0-crossing
        if self._edge_method == Edge.FWHM:
            raise ValueError(
                "FWHM edge method does not have inflection points. Use a different edge detection method"
            )
        d1 = np.gradient(
            gaussian_filter1d(
                self.values, sigma=self._edge_smoothing_ratio * len(self.values)
            )
        )
        (peak_idxs, _) = MultiProfile(d1).find_peaks(threshold=0.8)
        (valley_idxs, _) = MultiProfile(d1).find_valleys(threshold=0.8)
        left_idx = self._x_interp_to_original(peak_idxs[0])  # left-most index
        right_idx = self._x_interp_to_original(valley_idxs[-1])  # right-most index
        if self._edge_method == Edge.INFLECTION_DERIVATIVE:
            data = {
                "left index (rounded)": int(round(left_idx)),
                "left index (exact)": left_idx,
                "right index (rounded)": int(round(right_idx)),
                "right index (exact)": right_idx,
                "left value (@rounded)": self._y_original_to_interp(
                    int(round(left_idx))
                ),
                "left value (@exact)": self._y_original_to_interp(left_idx),
                "right value (@rounded)": self._y_original_to_interp(
                    int(round(right_idx))
                ),
                "right value (@exact)": self._y_original_to_interp(right_idx),
            }
            return data
        else:  # Hill
            # the 2nd deriv is a good approximation for the inflection point. Start there and fit Hill about it
            # penum_half_window = self.field_data()['width (exact)'] * self._hill_window_ratio / 2
            penum_half_window = int(
                round(self._hill_window_ratio * abs(right_idx - left_idx) / 2)
            )

            # left side
            x_data = np.array(
                [
                    x
                    for x in np.arange(
                        left_idx - penum_half_window, left_idx + penum_half_window
                    )
                    if x >= 0
                ]
            )
            y_data = self._y_original_to_interp(x_data)
            # y_data = self.values[left_idx - penum_half_window: left_idx + penum_half_window]
            left_hill = Hill.fit(x_data, y_data)
            left_infl = left_hill.inflection_idx()

            # right side
            x_data = np.array(
                [
                    x
                    for x in np.arange(
                        right_idx - penum_half_window, right_idx + penum_half_window
                    )
                    if x < len(d1)
                ]
            )
            y_data = self._y_original_to_interp(x_data)
            right_hill = Hill.fit(x_data, y_data)
            right_infl = right_hill.inflection_idx()

            data = {
                "left index (rounded)": left_infl["index (rounded)"],
                "left index (exact)": left_infl["index (exact)"],
                "right index (rounded)": right_infl["index (rounded)"],
                "right index (exact)": right_infl["index (exact)"],
                "left value (@exact)": left_hill.y(left_infl["index (exact)"]),
                "right value (@exact)": right_hill.y(right_infl["index (exact)"]),
                "left Hill params": left_hill.params,
                "right Hill params": right_hill.params,
            }
            return data

    def penumbra(self, lower: int = 20, upper: int = 80):
        """Calculate the penumbra of the field. Dependent on the edge detection method.

        Parameters
        ----------
        lower
            The lower % of the beam to use. If the edge method is FWHM, this is the typical % penumbra you're thinking.
            If the inflection method is used it will be the value/50 of the inflection point value. E.g. if the inflection
            point is perfectly at 50% with a ``lower`` of 20, then the penumbra value here will be 20% of the maximum.
            If the inflection point is at 30% of the max value (say for a FFF beam) then the lower penumbra will be ``lower/50``
            of the inflection point or ``0.3*lower/50``.
        upper
            Upper % of the beam to use. See lower for details.
        """
        if lower > upper:
            raise ValueError(
                "Upper penumbra value must be larger than the lower penumbra value"
            )
        if self._edge_method == Edge.FWHM:
            upper_data = self.fwxm_data(x=upper)
            lower_data = self.fwxm_data(x=lower)
            data = {
                f"left {lower}% index (exact)": lower_data["left index (exact)"],
                f"left {lower}% value (@rounded)": lower_data["left value (@rounded)"],
                f"left {upper}% index (exact)": upper_data["left index (exact)"],
                f"left {upper}% value (@rounded)": upper_data["left value (@rounded)"],
                f"right {lower}% index (exact)": lower_data["right index (exact)"],
                f"right {lower}% value (@rounded)": lower_data[
                    "right value (@rounded)"
                ],
                f"right {upper}% index (exact)": upper_data["right index (exact)"],
                f"right {upper}% value (@rounded)": upper_data[
                    "right value (@rounded)"
                ],
                "left values": self.values[
                    lower_data["left index (rounded)"] : upper_data[
                        "left index (rounded)"
                    ]
                ],
                "right values": self.values[
                    upper_data["right index (rounded)"] : lower_data[
                        "right index (rounded)"
                    ]
                ],
                f"left penumbra width (exact)": abs(
                    upper_data["left index (exact)"] - lower_data["left index (exact)"]
                ),
                f"right penumbra width (exact)": abs(
                    upper_data["right index (exact)"]
                    - lower_data["right index (exact)"]
                ),
            }
            if self.dpmm:
                data["left penumbra width (exact) mm"] = (
                    data["left penumbra width (exact)"] / self.dpmm
                )
                data["right penumbra width (exact) mm"] = (
                    data["right penumbra width (exact)"] / self.dpmm
                )
            return data
        elif self._edge_method == Edge.INFLECTION_DERIVATIVE:
            infl_data = self.inflection_data()
            lower_left_percent = max(
                infl_data["left value (@exact)"] / self.values.max() * lower / 50 * 100,
                1,
            )
            upper_left_percent = min(
                infl_data["left value (@exact)"] / self.values.max() * upper / 50 * 100,
                99,
            )
            upper_left_data = self.fwxm_data(x=upper_left_percent)
            lower_left_data = self.fwxm_data(x=lower_left_percent)

            lower_right_value = max(
                infl_data["right value (@exact)"]
                / self.values.max()
                * lower
                / 50
                * 100,
                1,
            )
            upper_right_value = min(
                infl_data["right value (@exact)"]
                / self.values.max()
                * upper
                / 50
                * 100,
                99,
            )
            upper_right_data = self.fwxm_data(x=upper_right_value)
            lower_right_data = self.fwxm_data(x=lower_right_value)

            data = {
                f"left {lower}% index (exact)": lower_left_data["left index (exact)"],
                f"left {upper}% index (exact)": upper_left_data["left index (exact)"],
                f"right {lower}% index (exact)": lower_right_data[
                    "right index (exact)"
                ],
                f"right {upper}% index (exact)": upper_right_data[
                    "right index (exact)"
                ],
                "left values": self._y_original_to_interp(
                    np.arange(
                        lower_left_data["left index (rounded)"],
                        upper_left_data["left index (rounded)"],
                    )
                ),
                "right values": self._y_original_to_interp(
                    np.arange(
                        upper_right_data["right index (rounded)"],
                        lower_right_data["right index (rounded)"],
                    )
                ),
                f"left penumbra width (exact)": abs(
                    upper_left_data["left index (exact)"]
                    - lower_left_data["left index (exact)"]
                ),
                f"right penumbra width (exact)": abs(
                    upper_right_data["right index (exact)"]
                    - lower_right_data["right index (exact)"]
                ),
            }
            if self.dpmm:
                data["left penumbra width (exact) mm"] = (
                    data["left penumbra width (exact)"] / self.dpmm
                )
                data["right penumbra width (exact) mm"] = (
                    data["right penumbra width (exact)"] / self.dpmm
                )
            return data
        elif self._edge_method == Edge.INFLECTION_HILL:
            infl_data = self.inflection_data()
            left_hill = Hill.from_params(infl_data["left Hill params"])
            right_hill = Hill.from_params(infl_data["right Hill params"])

            lower_left_percent = infl_data["left value (@exact)"] * lower / 50
            lower_left_index = left_hill.x(lower_left_percent)
            upper_left_percent = infl_data["left value (@exact)"] * upper / 50
            upper_left_index = left_hill.x(upper_left_percent)

            lower_right_value = infl_data["right value (@exact)"] * lower / 50
            lower_right_index = right_hill.x(lower_right_value)
            upper_right_value = infl_data["right value (@exact)"] * upper / 50
            upper_right_index = right_hill.x(upper_right_value)

            data = {
                f"left {lower}% index (exact)": lower_left_index,
                f"left {lower}% value (exact)": lower_left_percent,
                f"left {upper}% index (exact)": upper_left_index,
                f"left {upper}% value (exact)": upper_left_percent,
                f"right {lower}% index (exact)": lower_right_index,
                f"right {lower}% value (exact)": lower_right_value,
                f"right {upper}% index (exact)": upper_right_index,
                f"right {upper}% value (exact)": upper_right_value,
                "left values": self.values[
                    int(round(lower_left_index)) : int(round(upper_left_index))
                ],
                "right values": self.values[
                    int(round(upper_right_index)) : int(round(lower_right_index))
                ],
                f"left penumbra width (exact)": abs(
                    upper_left_index - lower_left_index
                ),
                f"right penumbra width (exact)": abs(
                    upper_right_index - lower_right_index
                ),
                f"left gradient (exact)": left_hill.gradient_at(
                    infl_data["left index (exact)"]
                ),
                r"right gradient (exact)": right_hill.gradient_at(
                    infl_data["right index (exact)"]
                ),
            }
            if self.dpmm:
                data["left penumbra width (exact) mm"] = (
                    data["left penumbra width (exact)"] / self.dpmm
                )
                data["left gradient (exact) %/mm"] = (
                    data["left gradient (exact)"] * self.dpmm * 100
                )  # 100 to convert to %
                data["right penumbra width (exact) mm"] = (
                    data["right penumbra width (exact)"] / self.dpmm
                )
                data["right gradient (exact) %/mm"] = (
                    data["right gradient (exact)"] * self.dpmm * 100
                )
            return data

    @argue.options(calculation=("mean", "median", "max", "min", "area"))
    def field_calculation(
        self, in_field_ratio: float = 0.8, calculation: str = "mean"
    ) -> Union[float, Tuple[float, float]]:
        """Perform an operation on the field values of the profile.
        This function is useful for determining field symmetry and flatness.

        Parameters
        ----------
        in_field_ratio
            Ratio of the field width to use in the calculation.
        calculation : {'mean', 'median', 'max', 'min', 'area'}
            Calculation to perform on the field values.
        """
        field_values = self.field_data(in_field_ratio)

        if calculation == "mean":
            return field_values["field values"].mean()
        elif calculation == "median":
            return float(np.median(field_values["field values"]))
        elif calculation == "max":
            return field_values["field values"].max()
        elif calculation == "min":
            return field_values["field values"].min()

    def gamma(
        self,
        evaluation_profile: SingleProfile,
        distance_to_agreement: float = 1,
        dose_to_agreement: float = 1,
        gamma_cap_value: float = 2,
        dose_threshold: float = 5,
        global_dose: bool = True,
        fill_value: float = np.nan,
    ) -> np.ndarray:
        """Calculate a 1D gamma. The passed profile is the evaluation profile. The instance calling this method is the reference profile.
        This profile must have the `dpmm` value given at instantiation so that physical spacing can be evaluated.
        The evaluation profile is resampled to be the same resolution as the reference profile.

        .. note::

            The difference between this method and the `gamma_1d` function is that 1) this is computed on Profile instances and 2)
            this validates the physical spacing of the profiles.

        Parameters
        ----------
        evaluation_profile
            The evaluation profile. This profile must have the `dpmm` value given at instantiation so that physical spacing can be evaluated.
        distance_to_agreement
            Distance in **mm** to search
        dose_to_agreement
            Dose in % of either global or local reference dose
        gamma_cap_value
            The value to cap the gamma at. E.g. a gamma of 5.3 will get capped to 2. Useful for displaying data with a consistent range.
        global_dose
            Whether to evaluate the dose to agreement threshold based on the global max or the dose point under evaluation.
        dose_threshold
            The dose threshold as a number between 0 and 100 of the % of max dose under which a gamma is not calculated.
            This is not affected by the global/local dose normalization and the threshold value is evaluated against the global max dose, period.
        fill_value
            The value to give pixels that were not calculated because they were under the dose threshold. Default
            is NaN, but another option would be 0. If NaN, allows the user to calculate mean/median gamma over just the
            evaluated portion and not be skewed by 0's that should not be considered.
        """
        if not self.dpmm or not evaluation_profile.dpmm:
            raise ValueError(
                "At least one profile does not have the dpmm attribute. Physical spacing cannot be determined. Set it before performing gamma analysis."
            )
        distance_to_agreement_px = int(round(distance_to_agreement * self.dpmm))
        # resample eval profile to be same resolution as reference
        resampled_evaluation = evaluation_profile.resample(
            interpolation_resolution_mm=self._interpolation_res
        )
        if len(resampled_evaluation.values) != len(self.values):
            warnings.warn(
                f"The number of elements in the reference and evaluation differ. Ref: {len(self.values)}, Eval: {len(resampled_evaluation.values)}"
            )
        # now that we've resampled, it's still possible that the x-values of the two profiles differ.
        # E.g. we may be at -0.475 and -0.37 for the first index depending on the amount of interpolation.
        # we thus need to evaluate the evaluation profile at the exact same x-indices as the reference.
        eval_at_ref_points = resampled_evaluation._y_original_to_interp(self.x_indices)
        return gamma_1d(
            reference=self.values,
            evaluation=eval_at_ref_points,
            dose_to_agreement=dose_to_agreement,
            distance_to_agreement=distance_to_agreement_px,
            gamma_cap_value=gamma_cap_value,
            global_dose=global_dose,
            dose_threshold=dose_threshold,
            fill_value=fill_value,
        )

    def plot(self, show: bool = True) -> None:
        """Plot the profile."""
        plt.plot(self.x_indices, self.values)
        if show:
            plt.show()


class MultiProfile(ProfileMixin):
    """A class for analyzing 1-D profiles that contain multiple signals. Methods are mostly for *finding & filtering*
    the signals, peaks, valleys, etc. Profiles with a single peak (e.g. radiation beam profiles) are better suited by the SingleProfile class.

    Attributes
    ----------
    values : ndarray
        The array of values passed in on instantiation.
    peaks : list
        List of Points, containing value and index information.
    valleys : list
        Same as peaks, but for valleys.

    """

    values: Union[np.ndarray, Sequence]
    peaks: List
    valleys: List

    def __init__(self, values: Union[np.ndarray, Sequence]):
        """
        Parameters
        ----------
        values : iterable
            Array of profile values.
        """
        self.values = values
        self.peaks = []
        self.valleys = []

    def plot(self, ax: Optional[plt.Axes] = None) -> None:
        """Plot the profile.

        Parameters
        ----------
        ax: plt.Axes
            An axis to plot onto. Optional.
        """
        if ax is None:
            fig, ax = plt.subplots()
        ax.plot(self.values)
        peaks_x = [peak.idx for peak in self.peaks]
        peaks_y = [peak.value for peak in self.peaks]
        ax.plot(peaks_x, peaks_y, "gv")
        valley_x = [peak.idx for peak in self.valleys]
        valley_y = [peak.value for peak in self.valleys]
        ax.plot(valley_x, valley_y, "r^")

    def find_peaks(
        self,
        threshold: Union[float, int] = 0.3,
        min_distance: Union[float, int] = 0.05,
        max_number: int = None,
        search_region: Tuple = (0.0, 1.0),
        peak_sort="prominences",
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Find the peaks of the profile using a simple maximum value search. This also sets the `peaks` attribute.

        Parameters
        ----------
        threshold : int, float
            The value the peak must be above to be considered a peak. This removes "peaks"
            that are in a low-value region.
            If passed an int, the actual value is the threshold.
            E.g. when passed 15, any peak less with a value <15 is removed.
            If passed a float, it will threshold as a percent. Must be between 0 and 1.
            E.g. when passed 0.4, any peak <40% of the maximum value will be removed.
        min_distance : int, float
            If passed an int, parameter is the number of elements apart a peak must be from neighboring peaks.
            If passed a float, must be between 0 and 1 and represents the ratio of the profile to exclude.
            E.g. if passed 0.05 with a 1000-element profile, the minimum peak width will be 0.05*1000 = 50 elements.
        max_number : int, None
            Specify up to how many peaks will be returned. E.g. if 3 is passed in and 5 peaks are found, only the 3 largest
            peaks will be returned. If None, no limit will be applied.
        search_region : tuple of ints, floats, or both
            The region within the profile to search. The tuple specifies the (left, right) edges to search.
            This allows exclusion of edges from the search. If a value is an int, it is taken as is. If a float, must
            be between 0 and 1 and is the ratio of the profile length. The left value must be less than the right.

        Returns
        -------
        indices: ndarray, values, ndarray
            The indices and values of the peaks.
        """
        peak_idxs, peak_props = find_peaks(
            self.values,
            threshold=threshold,
            peak_separation=min_distance,
            max_number=max_number,
            search_region=search_region,
            peak_sort=peak_sort,
        )
        self.peaks = [
            Point(value=peak_val, idx=peak_idx)
            for peak_idx, peak_val in zip(peak_idxs, peak_props["peak_heights"])
        ]

        return peak_idxs, peak_props["peak_heights"]

    def find_valleys(
        self,
        threshold: Union[float, int] = 0.3,
        min_distance: Union[float, int] = 0.05,
        max_number: int = None,
        search_region: Tuple = (0.0, 1.0),
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Find the valleys (minimums) of the profile using a simple minimum value search.

        Returns
        -------
        indices: ndarray, values, ndarray
            The indices and values of the valleys.

        See Also
        --------
        :meth:`~pylinac.core.profile.MultiProfile.find_peaks` : Further parameter info.
        """
        valley_idxs, valley_props = find_peaks(
            -self.values,
            threshold=threshold,
            peak_separation=min_distance,
            max_number=max_number,
            search_region=search_region,
        )
        self.valleys = [
            Point(value=self.values[valley_idx], idx=valley_idx)
            for valley_idx, valley_val in zip(
                valley_idxs, -valley_props["peak_heights"]
            )
        ]

        return valley_idxs, self.values[valley_idxs]

    def find_fwxm_peaks(
        self,
        threshold: Union[float, int] = 0.3,
        min_distance: Union[float, int] = 0.05,
        max_number: int = None,
        search_region: Tuple = (0.0, 1.0),
        peak_sort: str = "prominences",
        required_prominence=None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Find peaks using the center of the FWXM (rather than by max value).

        Parameters
        ----------
        x : int, float
            The Full-Width-X-Maximum desired. E.g. 0.7 will return the FW70%M.
            Values must be between 0 and 100.

        See Also
        --------
        find_peaks : Further parameter info
        """
        _, peak_props = find_peaks(
            self.values,
            threshold=threshold,
            min_width=min_distance,
            max_number=max_number,
            search_region=search_region,
            peak_sort=peak_sort,
            required_prominence=required_prominence,
        )
        fwxm_peak_idxs = []
        for lt, rt in zip(peak_props["left_ips"], peak_props["right_ips"]):
            fwxm = int(round(lt + (rt - lt) / 2))
            fwxm_peak_idxs.append(fwxm)

        fwxm_peak_vals = [self.values[fwxm] for fwxm in fwxm_peak_idxs]
        self.peaks = [
            Point(value=peak_val, idx=peak_idx)
            for peak_idx, peak_val in zip(fwxm_peak_idxs, fwxm_peak_vals)
        ]

        return np.array(fwxm_peak_idxs), np.array(fwxm_peak_vals)


class CircleProfile(MultiProfile, Circle):
    """A profile in the shape of a circle.

    Attributes
    ----------
    image_array : ndarray
        The 2D image array.
    start_angle : int, float
        Starting position of the profile in radians; 0 is right (0 on unit circle).
    ccw : bool
        How the profile is/was taken; clockwise or counter-clockwise.
    """

    image_array: np.ndarray
    start_angle: Union[float, int]
    ccw: bool
    sampling_ratio: float
    _x_locations: Optional[np.ndarray]
    _y_locations: Optional[np.ndarray]

    def __init__(
        self,
        center: Point,
        radius: float,
        image_array: np.ndarray,
        start_angle: Union[float, int] = 0,
        ccw: bool = True,
        sampling_ratio: float = 1.0,
    ):
        """
        Parameters
        ----------
        image_array : ndarray
            The 2D image array.
        start_angle : int, float
            Starting position of the profile in radians; 0 is right (0 on unit circle).
        ccw : bool
            If True (default), the profile will proceed counter-clockwise (the direction on the unit circle).
            If False, will proceed clockwise.
        sampling_ratio : float
            The ratio of pixel sampling to real pixels. E.g. if 1.0, the profile will have approximately
            the same number of elements as was encountered in the profile. A value of 2.0 will sample
            the profile at 2x the number of elements.

        See Also
        --------
        :class:`~pylinac.core.geometry.Circle` : Further parameter info.
        """
        Circle.__init__(self, center, radius)
        self._ensure_array_size(
            image_array, self.radius + self.center.x, self.radius + self.center.y
        )
        self.image_array = image_array
        self.start_angle = start_angle
        self.ccw = ccw
        self.sampling_ratio = sampling_ratio
        self._x_locations = None
        self._y_locations = None
        MultiProfile.__init__(self, self._profile)

    @property
    def size(self) -> float:
        """The elemental size of the profile."""
        return np.pi * self.radius * 2 * self.sampling_ratio

    @property
    def _radians(self) -> np.ndarray:
        interval = (2 * np.pi) / self.size
        rads = np.arange(
            0 + self.start_angle, (2 * np.pi) + self.start_angle - interval, interval
        )
        if self.ccw:
            rads = rads[::-1]
        return rads

    @property
    def x_locations(self) -> np.ndarray:
        """The x-locations of the profile values."""
        if self._x_locations is None:
            return np.cos(self._radians) * self.radius + self.center.x
        else:
            return self._x_locations

    @x_locations.setter
    def x_locations(self, array: np.ndarray):
        self._x_locations = array

    @property
    def y_locations(self) -> np.ndarray:
        """The x-locations of the profile values."""
        if self._y_locations is None:
            return np.sin(self._radians) * self.radius + self.center.y
        else:
            return self._y_locations

    @y_locations.setter
    def y_locations(self, array: np.ndarray):
        self._y_locations = array

    @property
    def _profile(self) -> np.ndarray:
        """The actual profile array; private attr that is passed to MultiProfile."""
        return ndimage.map_coordinates(
            self.image_array, [self.y_locations, self.x_locations], order=0
        )

    def find_peaks(
        self,
        threshold: Union[float, int] = 0.3,
        min_distance: Union[float, int] = 0.05,
        max_number: int = None,
        search_region: Tuple[float, float] = (0.0, 1.0),
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Overloads Profile to also map peak locations to the image."""
        peak_idxs, peak_vals = super().find_peaks(
            threshold, min_distance, max_number, search_region
        )
        self._map_peaks()
        return peak_idxs, peak_vals

    def find_valleys(
        self,
        threshold: Union[float, int] = 0.3,
        min_distance: Union[float, int] = 0.05,
        max_number: int = None,
        search_region: Tuple[float, float] = (0.0, 1.0),
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Overload Profile to also map valley locations to the image."""
        valley_idxs, valley_vals = super().find_valleys(
            threshold, min_distance, max_number, search_region
        )
        self._map_peaks()
        return valley_idxs, valley_vals

    def find_fwxm_peaks(
        self,
        threshold: Union[float, int] = 0.3,
        min_distance: Union[float, int] = 0.05,
        max_number: int = None,
        search_region: Tuple[float, float] = (0.0, 1.0),
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Overloads Profile to also map the peak locations to the image."""
        peak_idxs, peak_vals = super().find_fwxm_peaks(
            threshold, min_distance, max_number, search_region=search_region
        )
        self._map_peaks()
        return peak_idxs, peak_vals

    def _map_peaks(self) -> None:
        """Map found peaks to the x,y locations on the image/array; i.e. adds x,y coordinates to the peak locations"""
        for peak in self.peaks:
            peak.x = self.x_locations[int(peak.idx)]
            peak.y = self.y_locations[int(peak.idx)]

    def roll(self, amount: int) -> None:
        """Roll the profile and x and y coordinates."""
        self.values = np.roll(self.values, -amount)
        self.x_locations = np.roll(self.x_locations, -amount)
        self.y_locations = np.roll(self.y_locations, -amount)

    def plot2axes(
        self,
        axes: plt.Axes = None,
        edgecolor: str = "black",
        fill: bool = False,
        plot_peaks: bool = True,
    ) -> None:
        """Plot the circle to an axes.

        Parameters
        ----------
        axes : matplotlib.Axes, None
            The axes to plot on. If None, will create a new figure of the image array.
        edgecolor : str
            Color of the Circle; must be a valid matplotlib color.
        fill : bool
            Whether to fill the circle. matplotlib keyword.
        plot_peaks : bool
            If True, plots the found peaks as well.
        """
        if axes is None:
            fig, axes = plt.subplots()
            axes.imshow(self.image_array)
        axes.add_patch(
            mpl_Circle(
                (self.center.x, self.center.y),
                edgecolor=edgecolor,
                radius=self.radius,
                fill=fill,
            )
        )
        if plot_peaks:
            x_locs = [peak.x for peak in self.peaks]
            y_locs = [peak.y for peak in self.peaks]
            axes.autoscale(enable=False)
            axes.scatter(x_locs, y_locs, s=40, marker="x", c=edgecolor)

    @staticmethod
    def _ensure_array_size(
        array: np.ndarray, min_width: float, min_height: float
    ) -> None:
        """Ensure the array size of inputs are greater than the minimums."""
        height = array.shape[0]
        width = array.shape[1]
        if width < min_width or height < min_height:
            raise ValueError("Array size not large enough to compute profile")


class CollapsedCircleProfile(CircleProfile):
    """A circular profile that samples a thick band around the nominal circle, rather than just a 1-pixel-wide profile
    to give a mean value.
    """

    width_ratio: float
    num_profiles: int

    @argue.bounds(width_ratio=(0, 1))
    def __init__(
        self,
        center: Point,
        radius: float,
        image_array: Union[np.ndarray, "ArrayImage"],
        start_angle: int = 0,
        ccw: bool = True,
        sampling_ratio: float = 1.0,
        width_ratio: float = 0.1,
        num_profiles: int = 20,
    ):
        """
        Parameters
        ----------
        width_ratio : float
            The "thickness" of the band to sample. The ratio is relative to the radius. E.g. if the radius is 20
            and the width_ratio is 0.2, the "thickness" will be 4 pixels.
        num_profiles : int
            The number of profiles to sample in the band. Profiles are distributed evenly within the band.

        See Also
        --------
        :class:`~pylinac.core.profile.CircleProfile` : Further parameter info.
        """
        self.width_ratio = width_ratio
        self.num_profiles = num_profiles
        super().__init__(center, radius, image_array, start_angle, ccw, sampling_ratio)

    @property
    def _radii(self) -> np.ndarray:
        return np.linspace(
            start=self.radius * (1 - self.width_ratio),
            stop=self.radius * (1 + self.width_ratio),
            num=self.num_profiles,
        )

    @property
    def size(self) -> float:
        return np.pi * max(self._radii) * 2 * self.sampling_ratio

    @property
    def _multi_x_locations(self) -> List:
        """List of x-locations of the sampling profiles"""
        x = []
        cos = np.cos(self._radians)
        # extract profile for each circle radii
        for radius in self._radii:
            x.append(cos * radius + self.center.x)
        return x

    @property
    def _multi_y_locations(self) -> List:
        """List of x-locations of the sampling profiles"""
        y = []
        sin = np.sin(self._radians)
        # extract profile for each circle radii
        for radius in self._radii:
            y.append(sin * radius + self.center.y)
        return y

    @property
    def _profile(self) -> np.ndarray:
        """The actual profile array; private attr that is passed to MultiProfile."""
        profile = np.zeros(len(self._multi_x_locations[0]))
        for radius, x, y in zip(
            self._radii, self._multi_x_locations, self._multi_y_locations
        ):
            profile += ndimage.map_coordinates(self.image_array, [y, x], order=0)
        profile /= self.num_profiles
        return profile

    def plot2axes(
        self,
        axes: plt.Axes = None,
        edgecolor: str = "black",
        fill: bool = False,
        plot_peaks: bool = True,
    ) -> None:
        """Add 2 circles to the axes: one at the maximum and minimum radius of the ROI.

        See Also
        --------
        :meth:`~pylinac.core.profile.CircleProfile.plot2axes` : Further parameter info.
        """
        if axes is None:
            fig, axes = plt.subplots()
            axes.imshow(self.image_array)
        axes.add_patch(
            mpl_Circle(
                (self.center.x, self.center.y),
                edgecolor=edgecolor,
                radius=self.radius * (1 + self.width_ratio),
                fill=fill,
            )
        )
        axes.add_patch(
            mpl_Circle(
                (self.center.x, self.center.y),
                edgecolor=edgecolor,
                radius=self.radius * (1 - self.width_ratio),
                fill=fill,
            )
        )
        if plot_peaks:
            x_locs = [peak.x for peak in self.peaks]
            y_locs = [peak.y for peak in self.peaks]
            axes.autoscale(enable=False)
            axes.scatter(x_locs, y_locs, s=20, marker="x", c=edgecolor)


def find_peaks(
    values: np.ndarray,
    threshold: Union[float, int] = -np.inf,
    peak_separation: Union[float, int] = 0,
    max_number: int = None,
    fwxm_height: float = 0.5,
    min_width: int = 0,
    search_region: Tuple[float, float] = (0.0, 1.0),
    peak_sort="prominences",
    required_prominence=None,
) -> Tuple[np.ndarray, dict]:
    """Find the peaks of a 1D signal. Heavily relies on the scipy implementation.

    Parameters
    ----------
    values : array-like
        Signal values to search for peaks within.
    threshold : int, float
        The value the peak must be above to be considered a peak. This removes "peaks"
        that are in a low-value region.
        If passed an int, the actual value is the threshold.
        E.g. when passed 15, any peak less with a value <15 is removed.
        If passed a float, it will threshold as a percent. Must be between 0 and 1.
        E.g. when passed 0.4, any peak <40% of the maximum value will be removed.
    peak_separation : int, float
        If passed an int, parameter is the number of elements apart a peak must be from neighboring peaks.
        If passed a float, must be between 0 and 1 and represents the ratio of the profile to exclude.
        E.g. if passed 0.05 with a 1000-element profile, the minimum peak width will be 0.05*1000 = 50 elements.
    max_number : int, None
        Specify up to how many peaks will be returned. E.g. if 3 is passed in and 5 peaks are found, only the 3 largest
        peaks will be returned.
    fwxm_height: float
        The relative height at which a FWXM calculation is performed. Although this function finds simple max values,
        the underlying function can provide fwxm information as well.
    min_width: int
        The minimum width of the peak.
    search_region: tuple
        The search region to use within the values.
        Using between 0 and 1 will convert to a ratio of the indices. E.g. to search the middle half of the passed values, use (0.25, 0.75).
        Using ints above 1 will use the indices directly. E.g. (33, 71) will search between those two indices.

    Returns
    -------
    peak_idxs : numpy.array
        The indices of the peaks found.
    peak_props : dict
        A dict containing contextual peak data.
    """
    peak_separation, shift_amount, threshold, trimmed_values = _parse_peak_args(
        peak_separation, search_region, threshold, values
    )

    peak_idxs, peak_props = signal.find_peaks(
        trimmed_values,
        rel_height=(1 - fwxm_height),
        width=min_width,
        height=threshold,
        distance=peak_separation,
        prominence=required_prominence,
    )
    peak_idxs += shift_amount  # shift according to the search region left edge

    # get the "largest" peaks up to max number, and then re-sort to be left->right like it was originally
    largest_peak_idxs = sorted(
        list(np.argsort(peak_props[peak_sort]))[::-1][:max_number]
    )

    # cut down prop arrays as need be
    for key, array_vals in peak_props.items():
        peak_props[key] = array_vals[largest_peak_idxs]
    return peak_idxs[largest_peak_idxs], peak_props


def _parse_peak_args(
    peak_separation: float,
    search_region: Tuple[float, float],
    threshold: float,
    values: np.ndarray,
) -> Tuple[float, int, float, np.ndarray]:
    """Converts arguments as needed. E.g. converting a ratio to actual values"""
    # set threshold as % if between 0 and 1
    val_range = values.max() - values.min()
    if 0 <= threshold <= 1:
        threshold = values.min() + threshold * val_range
    # set separation as % if between 0 and 1
    if 0 <= peak_separation <= 1:
        peak_separation = max(int(peak_separation * len(values)), 1)
    # limit to search region
    if max(search_region) <= 1:
        shift_amount = int(search_region[0] * len(values))
        values = values[
            int(search_region[0] * len(values)) : int(search_region[1] * len(values))
        ]
    else:
        values = values[search_region[0] : search_region[1]]
        shift_amount = search_region[0]
    return peak_separation, shift_amount, threshold, values



"""This module holds classes for image loading and manipulation."""
import copy
import io
import math
import os
import os.path as osp
import re
from collections import Counter
from datetime import datetime
from io import BytesIO, BufferedReader
from pathlib import Path
from typing import Union, Sequence, List, Any, Tuple, Optional, BinaryIO

import argue
import matplotlib.pyplot as plt
import numpy as np
import pydicom
import scipy.ndimage.filters as spf
from PIL import Image as pImage
from pydicom.errors import InvalidDicomError
from scipy import ndimage
from skimage.draw import disk

from .geometry import Point
from .io import (
    get_url,
    TemporaryZipDirectory,
    retrieve_filenames,
    is_dicom_image,
    retrieve_dicom_file,
)
from .profile import stretch as stretcharray
from .utilities import is_close
from ..settings import get_dicom_cmap, PATH_TRUNCATION_LENGTH

ARRAY = "Array"
DICOM = "DICOM"
IMAGE = "Image"

FILE_TYPE = "file"
STREAM_TYPE = "stream"

MM_PER_INCH = 25.4

ImageLike = Union["DicomImage", "ArrayImage", "FileImage", "LinacDicomImage"]


def equate_images(image1: ImageLike, image2: ImageLike) -> Tuple[ImageLike, ImageLike]:
    """Crop and resize two images to make them:
      * The same pixel dimensions
      * The same DPI

    The usefulness of the function comes when trying to compare images from different sources.
    The best example is calculating gamma on a machine log fluence and EPID image. The physical
    and pixel dimensions must be normalized, the SID normalized

    Parameters
    ----------
    image1 : {:class:`~pylinac.core.image.ArrayImage`, :class:`~pylinac.core.image.DicomImage`, :class:`~pylinac.core.image.FileImage`}
        Must have DPI and SID.
    image2 : {:class:`~pylinac.core.image.ArrayImage`, :class:`~pylinac.core.image.DicomImage`, :class:`~pylinac.core.image.FileImage`}
        Must have DPI and SID.

    Returns
    -------
    image1 : :class:`~pylinac.core.image.ArrayImage`
    image2 : :class:`~pylinac.core.image.ArrayImage`
        The returns are new instances of Images.
    """
    image1 = copy.deepcopy(image1)
    image2 = copy.deepcopy(image2)
    # crop images to be the same physical size
    # ...crop height
    physical_height_diff = image1.physical_shape[0] - image2.physical_shape[0]
    if physical_height_diff < 0:  # image2 is bigger
        img = image2
    else:
        img = image1
    pixel_height_diff = abs(int(round(-physical_height_diff * img.dpmm / 2)))
    img.crop(pixel_height_diff, edges=("top", "bottom"))

    # ...crop width
    physical_width_diff = image1.physical_shape[1] - image2.physical_shape[1]
    if physical_width_diff > 0:
        img = image1
    else:
        img = image2
    pixel_width_diff = abs(int(round(physical_width_diff * img.dpmm / 2)))
    img.crop(pixel_width_diff, edges=("left", "right"))

    # resize images to be of the same shape
    zoom_factor = image1.shape[1] / image2.shape[1]
    image2_array = ndimage.interpolation.zoom(image2.as_type(float), zoom_factor)
    image2 = load(image2_array, dpi=image2.dpi * zoom_factor)

    return image1, image2


def is_image(path: Union[str, io.BytesIO, ImageLike, np.ndarray]) -> bool:
    """Determine whether the path is a valid image file.

    Returns
    -------
    bool
    """
    return any((_is_array(path), _is_dicom(path), _is_image_file(path)))


def retrieve_image_files(path: str) -> List[str]:
    """Retrieve the file names of all the valid image files in the path.

    Returns
    -------
    list
        Contains strings pointing to valid image paths.
    """
    return retrieve_filenames(directory=path, func=is_image)


def load(
    path: Union[str, Path, ImageLike, np.ndarray, BinaryIO], **kwargs
) -> ImageLike:
    """Load a DICOM image, JPG/TIF/BMP image, or numpy 2D array.

    Parameters
    ----------
    path : str, file-object
        The path to the image file or data stream or array.
    kwargs
        See :class:`~pylinac.core.image.FileImage`, :class:`~pylinac.core.image.DicomImage`,
        or :class:`~pylinac.core.image.ArrayImage` for keyword arguments.

    Returns
    -------
    ::class:`~pylinac.core.image.FileImage`, :class:`~pylinac.core.image.ArrayImage`, or :class:`~pylinac.core.image.DicomImage`
        Return type depends on input image.

    Examples
    --------
    Load an image from a file and then apply a filter::

        >>> from pylinac.core.image import load
        >>> my_image = r"C:\QA\image.tif"
        >>> img = load(my_image)  # returns a FileImage
        >>> img.filter(5)

    Loading from an array is just like loading from a file::

        >>> arr = np.arange(36).reshape(6, 6)
        >>> img = load(arr)  # returns an ArrayImage
    """
    if isinstance(path, BaseImage):
        return path

    if _is_array(path):
        return ArrayImage(path, **kwargs)
    elif _is_dicom(path):
        return DicomImage(path, **kwargs)
    elif _is_image_file(path):
        return FileImage(path, **kwargs)
    else:
        raise TypeError(
            f"The argument `{path}` was not found to be a valid DICOM file, Image file, or array"
        )


def load_url(url: str, progress_bar: bool = True, **kwargs) -> ImageLike:
    """Load an image from a URL.

    Parameters
    ----------
    url : str
        A string pointing to a valid URL that points to a file.

        .. note:: For some images (e.g. Github), the raw binary URL must be used, not simply the basic link.

    progress_bar: bool
        Whether to display a progress bar of download status.
    """
    filename = get_url(url, progress_bar=progress_bar)
    return load(filename, **kwargs)


def load_multiples(
    image_file_list: Sequence,
    method: str = "mean",
    stretch_each: bool = True,
    **kwargs,
) -> ImageLike:
    """Combine multiple image files into one superimposed image.

    Parameters
    ----------
    image_file_list : list
        A list of the files to be superimposed.
    method : {'mean', 'max', 'sum'}
        A string specifying how the image values should be combined.
    stretch_each : bool
        Whether to normalize the images being combined by stretching their high/low values to the same values across images.
    kwargs :
        Further keyword arguments are passed to the load function and stretch function.

    Examples
    --------
    Load multiple images::

        >>> from pylinac.core.image import load_multiples
        >>> paths = ['starshot1.tif', 'starshot2.tif']
        >>> superimposed_img = load_multiples(paths)
    """
    # load images
    img_list = [load(path, **kwargs) for path in image_file_list]
    first_img = img_list[0]

    # check that all images are the same size and stretch if need be
    for img in img_list:
        if img.shape != first_img.shape:
            raise ValueError("Images were not the same shape")
        if stretch_each:
            img.array = stretcharray(img.array, fill_dtype=kwargs.get("dtype"))

    # stack and combine arrays
    new_array = np.dstack(tuple(img.array for img in img_list))
    if method == "mean":
        combined_arr = np.mean(new_array, axis=2)
    elif method == "max":
        combined_arr = np.max(new_array, axis=2)
    elif method == "sum":
        combined_arr = np.sum(new_array, axis=2)

    # replace array of first object and return
    first_img.array = combined_arr
    return first_img


def _is_dicom(path: Union[str, Path, io.BytesIO, ImageLike, np.ndarray]) -> bool:
    """Whether the file is a readable DICOM file via pydicom."""
    return is_dicom_image(file=path)


def _is_image_file(path: Union[str, Path]) -> bool:
    """Whether the file is a readable image file via Pillow."""
    try:
        pImage.open(path)
        return True
    except:
        return False


def _is_array(obj: Any) -> bool:
    """Whether the object is a numpy array."""
    return isinstance(obj, np.ndarray)


class BaseImage:
    """Base class for the Image classes.

    Attributes
    ----------
    path : str
        The path to the image file.
    array : numpy.ndarray
        The actual image pixel array.
    """

    def __init__(
        self, path: Union[str, Path, BytesIO, ImageLike, np.ndarray, BufferedReader]
    ):
        """
        Parameters
        ----------
        path : str
            The path to the image.
        """
        source: Union[FILE_TYPE, STREAM_TYPE]

        if isinstance(path, (str, Path)) and not osp.isfile(path):
            raise FileExistsError(
                f"File `{path}` does not exist. Verify the file path name."
            )
        elif isinstance(path, (str, Path)) and osp.isfile(path):
            self.path = path
            self.base_path = osp.basename(path)
            self.source = FILE_TYPE
        else:
            self.source = STREAM_TYPE
            path.seek(0)
            try:
                self.path = str(Path(path.name))
            except AttributeError:
                self.path = ""

    @property
    def truncated_path(
        self,
    ) -> str:  # TODO: Use textwrap or pull out into util function
        if self.source == FILE_TYPE:
            path = str(self.path)
            if len(path) > PATH_TRUNCATION_LENGTH:
                return (
                    path[: PATH_TRUNCATION_LENGTH // 2]
                    + "..."
                    + path[-PATH_TRUNCATION_LENGTH // 2 :]
                )
            else:
                return path
        else:
            return ""  # was from stream, no path

    @classmethod
    def from_multiples(
        cls,
        filelist: List[str],
        method: str = "mean",
        stretch: bool = True,
        **kwargs,
    ) -> ImageLike:
        """Load an instance from multiple image items. See :func:`~pylinac.core.image.load_multiples`."""
        return load_multiples(filelist, method, stretch, **kwargs)

    @property
    def center(self) -> Point:
        """Return the center position of the image array as a Point.
        Even-length arrays will return the midpoint between central two indices. Odd will return the central index."""
        x_center = (self.shape[1] / 2) - 0.5
        y_center = (self.shape[0] / 2) - 0.5
        return Point(x_center, y_center)

    @property
    def physical_shape(self) -> (float, float):
        """The physical size of the image in mm."""
        return self.shape[0] / self.dpmm, self.shape[1] / self.dpmm

    def date_created(self, format: str = "%A, %B %d, %Y") -> str:
        """The date the file was created. Tries DICOM data before falling back on OS timestamp"""
        date = None
        try:
            date = datetime.strptime(
                self.metadata.InstanceCreationDate
                + str(round(float(self.metadata.InstanceCreationTime))),
                "%Y%m%d%H%M%S",
            )
            date = date.strftime(format)
        except (AttributeError, ValueError):
            try:
                date = datetime.strptime(self.metadata.StudyDate, "%Y%m%d")
                date = date.strftime(format)
            except:
                pass
        if date is None:
            try:
                date = datetime.fromtimestamp(osp.getctime(self.path)).strftime(format)
            except AttributeError:
                date = "Unknown"
        return date

    def plot(
        self, ax: plt.Axes = None, show: bool = True, clear_fig: bool = False, **kwargs
    ) -> plt.Axes:
        """Plot the image.

        Parameters
        ----------
        ax : matplotlib.Axes instance
            The axis to plot the image to. If None, creates a new figure.
        show : bool
            Whether to actually show the image. Set to false when plotting multiple items.
        clear_fig : bool
            Whether to clear the prior items on the figure before plotting.
        kwargs
            kwargs passed to plt.imshow()
        """
        if ax is None:
            fig, ax = plt.subplots()
        if clear_fig:
            plt.clf()
        ax.imshow(self.array, cmap=get_dicom_cmap(), **kwargs)
        if show:
            plt.show()
        return ax

    def filter(
        self,
        size: Union[float, int] = 0.05,
        kind: str = "median",
    ) -> None:
        """Filter the profile in place.

        Parameters
        ----------
        size : int, float
            Size of the median filter to apply.
            If a float, the size is the ratio of the length. Must be in the range 0-1.
            E.g. if size=0.1 for a 1000-element array, the filter will be 100 elements.
            If an int, the filter is the size passed.
        kind : {'median', 'gaussian'}
            The kind of filter to apply. If gaussian, *size* is the sigma value.
        """
        if isinstance(size, float):
            if 0 < size < 1:
                size *= len(self.array)
                size = max(size, 1)
            else:
                raise TypeError("Float was passed but was not between 0 and 1")

        if kind == "median":
            self.array = ndimage.median_filter(self.array, size=size)
        elif kind == "gaussian":
            self.array = ndimage.gaussian_filter(self.array, sigma=size)

    def crop(
        self,
        pixels: int = 15,
        edges: Tuple[str, ...] = ("top", "bottom", "left", "right"),
    ) -> None:
        """Removes pixels on all edges of the image in-place.

        Parameters
        ----------
        pixels : int
            Number of pixels to cut off all sides of the image.
        edges : tuple
            Which edges to remove from. Can be any combination of the four edges.
        """
        if pixels < 0:
            raise ValueError("Pixels to remove must be a positive number")
        if "top" in edges:
            self.array = self.array[pixels:, :]
        if "bottom" in edges:
            self.array = self.array[:-pixels, :]
        if "left" in edges:
            self.array = self.array[:, pixels:]
        if "right" in edges:
            self.array = self.array[:, :-pixels]

    def flipud(self) -> None:
        """Flip the image array upside down in-place. Wrapper for np.flipud()"""
        self.array = np.flipud(self.array)

    def fliplr(self) -> None:
        """Flip the image array upside down in-place. Wrapper for np.fliplr()"""
        self.array = np.fliplr(self.array)

    def invert(self) -> None:
        """Invert (imcomplement) the image."""
        orig_array = self.array
        self.array = -orig_array + orig_array.max() + orig_array.min()

    def roll(self, direction: str = "x", amount: int = 1) -> None:
        """Roll the image array around in-place. Wrapper for np.roll().

        Parameters
        ----------
        direction : {'x', 'y'}
            The axis to roll over.
        amount : int
            The amount of elements to roll over.
        """
        axis = 1 if direction == "x" else 0
        self.array = np.roll(self.array, amount, axis=axis)

    def rot90(self, n: int = 1) -> None:
        """Wrapper for numpy.rot90; rotate the array by 90 degrees CCW n times."""
        self.array = np.rot90(self.array, n)

    def threshold(
        self, threshold: float, kind: str = "high"
    ) -> None:
        """Apply a high- or low-pass threshold filter.

        Parameters
        ----------
        threshold : int
            The cutoff value.
        kind : str
            If ``high`` (default), will apply a high-pass threshold. All values above the cutoff are left as-is.
            Remaining points are set to 0.
            If ``low``, will apply a low-pass threshold.
        """
        if kind == "high":
            self.array = np.where(self.array >= threshold, self, 0)
        else:
            self.array = np.where(self.array <= threshold, self, 0)

    def as_binary(self, threshold: int) -> ImageLike:
        """Return a binary (black & white) image based on the given threshold.

        Parameters
        ----------
        threshold : int, float
            The threshold value. If the value is above or equal to the threshold it is set to 1, otherwise to 0.

        Returns
        -------
        ArrayImage
        """
        array = np.where(self.array >= threshold, 1, 0)
        return ArrayImage(array)

    def dist2edge_min(self, point: Union[Point, Tuple]) -> float:
        """Calculates distance from given point to the closest edge.

        Parameters
        ----------
        point : geometry.Point, tuple

        Returns
        -------
        float
        """
        if isinstance(point, tuple):
            point = Point(point)
        rows = self.shape[0]
        cols = self.shape[1]
        disttoedge = np.zeros(4)
        disttoedge[0] = rows - point.y
        disttoedge[1] = cols - point.x
        disttoedge[2] = point.y
        disttoedge[3] = point.x
        return min(disttoedge)

    def ground(self) -> float:
        """Ground the profile in place such that the lowest value is 0.

        .. note::
            This will also "ground" profiles that are negative or partially-negative.
            For such profiles, be careful that this is the behavior you desire.

        Returns
        -------
        float
            The amount subtracted from the image.
        """
        min_val = self.array.min()
        self.array -= min_val
        return min_val

    def normalize(self, norm_val: Union[str, float] = "max") -> None:
        """Normalize the image values in place to the given value.

        Parameters
        ----------
        norm_val : str, number
            If a string, must be 'max', which normalizes the values to the maximum value.
            If a number, normalizes all values to that number.
        """
        if norm_val == "max":
            val = self.array.max()
        else:
            val = norm_val
        self.array = self.array / val

    def check_inversion(
        self, box_size: int = 20, position: (float, float) = (0.0, 0.0)
    ) -> None:
        """Check the image for inversion by sampling the 4 image corners.
        If the average value of the four corners is above the average pixel value, then it is very likely inverted.

        Parameters
        ----------
        box_size : int
            The size in pixels of the corner box to detect inversion.
        position : 2-element sequence
            The location of the sampling boxes.
        """
        row_pos = max(int(position[0] * self.array.shape[0]), 1)
        col_pos = max(int(position[1] * self.array.shape[1]), 1)
        lt_upper = self.array[
            row_pos : row_pos + box_size, col_pos : col_pos + box_size
        ]
        rt_upper = self.array[
            row_pos : row_pos + box_size, -col_pos - box_size : -col_pos
        ]
        lt_lower = self.array[
            -row_pos - box_size : -row_pos, col_pos : col_pos + box_size
        ]
        rt_lower = self.array[
            -row_pos - box_size : -row_pos, -col_pos - box_size : -col_pos
        ]
        avg = np.mean((lt_upper, lt_lower, rt_upper, rt_lower))
        if avg > np.mean(self.array.flatten()):
            self.invert()

    def check_inversion_by_histogram(
        self, percentiles: (float, float, float) = (5, 50, 95)
    ) -> bool:
        """Check the inversion of the image using histogram analysis. The assumption is that the image
        is mostly background-like values and that there is a relatively small amount of dose getting to the image
        (e.g. a picket fence image). This function looks at the distance from one percentile to another to determine
        if the image should be inverted.

        Parameters
        ----------
        percentiles : 3-element tuple
            The 3 percentiles to compare. Default is (5, 50, 95). Recommend using (x, 50, y). To invert the other way
            (where pixel value is *decreasing* with dose, reverse the percentiles, e.g. (95, 50, 5).

        Returns
        -------
        bool: Whether an inversion was performed.
        """
        was_inverted = False
        p_low = np.percentile(self.array, percentiles[0])
        p_mid = np.percentile(self.array, percentiles[1])
        p_high = np.percentile(self.array, percentiles[2])
        mid_to_low = abs(p_mid - p_low)
        mid_to_high = abs(p_mid - p_high)
        if mid_to_low > mid_to_high:
            was_inverted = True
            self.invert()
        return was_inverted

    @argue.bounds(threshold=(0.0, 1.0))
    def gamma(
        self,
        comparison_image: ImageLike,
        doseTA: float = 1,
        distTA: float = 1,
        threshold: float = 0.1,
        ground: bool = True,
        normalize: bool = True,
    ) -> np.ndarray:
        """Calculate the gamma between the current image (reference) and a comparison image.

        .. versionadded:: 1.2

        The gamma calculation is based on `Bakai et al
        <http://iopscience.iop.org/0031-9155/48/21/006/>`_ eq.6,
        which is a quicker alternative to the standard Low gamma equation.

        Parameters
        ----------
        comparison_image : {:class:`~pylinac.core.image.ArrayImage`, :class:`~pylinac.core.image.DicomImage`, or :class:`~pylinac.core.image.FileImage`}
            The comparison image. The image must have the same DPI/DPMM to be comparable.
            The size of the images must also be the same.
        doseTA : int, float
            Dose-to-agreement in percent; e.g. 2 is 2%.
        distTA : int, float
            Distance-to-agreement in mm.
        threshold : float
            The dose threshold percentage of the maximum dose, below which is not analyzed.
            Must be between 0 and 1.
        ground : bool
            Whether to "ground" the image values. If true, this sets both datasets to have the minimum value at 0.
            This can fix offset errors in the data.
        normalize : bool
            Whether to normalize the images. This sets the max value of each image to the same value.

        Returns
        -------
        gamma_map : numpy.ndarray
            The calculated gamma map.

        See Also
        --------
        :func:`~pylinac.core.image.equate_images`
        """
        # error checking
        if not is_close(self.dpi, comparison_image.dpi, delta=0.1):
            raise AttributeError(
                f"The image DPIs to not match: {self.dpi:.2f} vs. {comparison_image.dpi:.2f}"
            )
        same_x = is_close(self.shape[1], comparison_image.shape[1], delta=1.1)
        same_y = is_close(self.shape[0], comparison_image.shape[0], delta=1.1)
        if not (same_x and same_y):
            raise AttributeError(
                f"The images are not the same size: {self.shape} vs. {comparison_image.shape}"
            )

        # set up reference and comparison images
        ref_img = ArrayImage(copy.copy(self.array))
        ref_img.check_inversion_by_histogram()
        if ground:
            ref_img.ground()
        if normalize:
            ref_img.normalize()
        comp_img = ArrayImage(copy.copy(comparison_image.array))
        comp_img.check_inversion_by_histogram()
        if ground:
            comp_img.ground()
        if normalize:
            comp_img.normalize()

        # invalidate dose values below threshold so gamma doesn't calculate over it
        ref_img.array[ref_img < threshold * np.max(ref_img)] = np.NaN

        # convert distance value from mm to pixels
        distTA_pixels = self.dpmm * distTA

        # construct image gradient using sobel filter
        img_x = spf.sobel(ref_img.as_type(np.float32), 1)
        img_y = spf.sobel(ref_img.as_type(np.float32), 0)
        grad_img = np.hypot(img_x, img_y)

        # equation: (measurement - reference) / sqrt ( doseTA^2 + distTA^2 * image_gradient^2 )
        subtracted_img = np.abs(comp_img - ref_img)
        denominator = np.sqrt(
            ((doseTA / 100.0) ** 2) + ((distTA_pixels ** 2) * (grad_img ** 2))
        )
        gamma_map = subtracted_img / denominator

        return gamma_map

    def as_type(self, dtype: np.dtype) -> np.ndarray:
        return self.array.astype(dtype)

    @property
    def shape(self) -> (int, int):
        return self.array.shape

    @property
    def size(self) -> int:
        return self.array.size

    @property
    def ndim(self) -> int:
        return self.array.ndim

    @property
    def dtype(self) -> np.dtype:
        return self.array.dtype

    def sum(self) -> float:
        return self.array.sum()

    def ravel(self) -> np.ndarray:
        return self.array.ravel()

    @property
    def flat(self) -> np.ndarray:
        return self.array.flat

    def __len__(self):
        return len(self.array)

    def __getitem__(self, item):
        return self.array[item]


class DicomImage(BaseImage):
    """An image from a DICOM RTImage file.

    Attributes
    ----------
    metadata : pydicom Dataset
        The dataset of the file as returned by pydicom without pixel data.
    """

    metadata: pydicom.FileDataset
    _sid: float
    _dpi: float

    def __init__(
        self,
        path: Union[str, Path, BytesIO, BufferedReader],
        *,
        dtype=None,
        dpi: float = None,
        sid: float = None,
    ):
        """
        Parameters
        ----------
        path : str, file-object
            The path to the file or the data stream.
        dtype : dtype, None, optional
            The data type to cast the image data as. If None, will use whatever raw image format is.
        dpi : int, float
            The dots-per-inch of the image, defined at isocenter.

            .. note:: If a DPI tag is found in the image, that value will override the parameter, otherwise this one
                will be used.

        sid : int, float
            The Source-to-Image distance in mm.
        """
        super().__init__(path)
        self._sid = sid
        self._dpi = dpi
        # read the file once to get just the DICOM metadata
        self.metadata = retrieve_dicom_file(path)
        self._original_dtype = self.metadata.pixel_array.dtype
        # read a second time to get pixel data
        try:
            path.seek(0)
        except AttributeError:
            pass
        ds = retrieve_dicom_file(path)
        if dtype is not None:
            self.array = ds.pixel_array.astype(dtype)
        else:
            self.array = ds.pixel_array.copy()
        # convert values to HU or CU: real_values = slope * raw + intercept
        has_all_rescale_tags = (
            hasattr(self.metadata, "RescaleSlope")
            and hasattr(self.metadata, "RescaleIntercept")
            and hasattr(self.metadata, "PixelIntensityRelationshipSign")
        )
        has_some_rescale_tags = hasattr(self.metadata, "RescaleSlope") and hasattr(
            self.metadata, "RescaleIntercept"
        )
        is_ct_storage = self.metadata.SOPClassUID.name == "CT Image Storage"
        is_mr_storage = self.metadata.SOPClassUID.name == "MR Image Storage"
        if has_all_rescale_tags:
            self.array = (
                (self.metadata.RescaleSlope * self.array)
                + self.metadata.RescaleIntercept
            ) * self.metadata.PixelIntensityRelationshipSign
        elif is_ct_storage or has_some_rescale_tags:
            self.array = (
                self.metadata.RescaleSlope * self.array
            ) + self.metadata.RescaleIntercept
        elif is_mr_storage:
            # signal is usually correct as-is, no inversion needed
            pass
        else:
            # invert it
            orig_array = self.array
            self.array = -orig_array + orig_array.max() + orig_array.min()

    def save(self, filename: Union[str, Path]) -> Union[str, Path]:
        """Save the image instance back out to a .dcm file.

        Returns
        -------
        A string pointing to the new filename.
        """
        if self.metadata.SOPClassUID.name == "CT Image Storage":
            self.array = (self.array - int(self.metadata.RescaleIntercept)) / int(
                self.metadata.RescaleSlope
            )
        self.metadata.PixelData = self.array.astype(self._original_dtype).tobytes()
        self.metadata.save_as(filename)
        return filename

    @property
    def sid(self) -> float:
        """The Source-to-Image in mm."""
        try:
            return float(self.metadata.RTImageSID)
        except:
            return self._sid

    @property
    def dpi(self) -> float:
        """The dots-per-inch of the image, defined at isocenter."""
        try:
            return self.dpmm * MM_PER_INCH
        except:
            return self._dpi

    @property
    def dpmm(self) -> float:
        """The Dots-per-mm of the image, defined at isocenter. E.g. if an EPID image is taken at 150cm SID,
        the dpmm will scale back to 100cm."""
        dpmm = None
        for tag in ("PixelSpacing", "ImagePlanePixelSpacing"):
            mmpd = self.metadata.get(tag)
            if mmpd is not None:
                dpmm = 1 / mmpd[0]
                break
        if dpmm is not None and self.sid is not None:
            dpmm *= self.sid / 1000
        elif dpmm is None and self._dpi is not None:
            dpmm = self._dpi / MM_PER_INCH
        return dpmm

    @property
    def cax(self) -> Point:
        """The position of the beam central axis. If no DICOM translation tags are found then the center is returned.
        Uses this tag: https://dicom.innolitics.com/ciods/rt-beams-delivery-instruction/rt-beams-delivery-instruction/00741020/00741030/3002000d"""
        try:
            x = self.center.x - self.metadata.XRayImageReceptorTranslation[0]
            y = self.center.y - self.metadata.XRayImageReceptorTranslation[1]
        except AttributeError:
            return self.center
        else:
            return Point(x, y)


class LinacDicomImage(DicomImage):
    """DICOM image taken on a linac. Also allows passing of gantry/coll/couch values via the filename."""

    gantry_keyword = "Gantry"
    collimator_keyword = "Coll"
    couch_keyword = "Couch"

    _use_filenames: bool

    def __init__(self, path: Union[str, Path], use_filenames: bool = False, **kwargs):
        self._gantry = kwargs.pop("gantry", None)
        self._coll = kwargs.pop("coll", None)
        self._couch = kwargs.pop("couch", None)
        super().__init__(path, **kwargs)
        self._use_filenames = use_filenames

    @property
    def gantry_angle(self) -> float:
        """Gantry angle of the irradiation."""
        if self._gantry is not None:
            return self._gantry
        else:
            return self._get_axis_value(self.gantry_keyword, "GantryAngle")

    @property
    def collimator_angle(self) -> float:
        """Collimator angle of the irradiation."""
        if self._coll is not None:
            return self._coll
        else:
            return self._get_axis_value(
                self.collimator_keyword, "BeamLimitingDeviceAngle"
            )

    @property
    def couch_angle(self) -> float:
        """Couch angle of the irradiation."""
        if self._couch is not None:
            return self._couch
        else:
            return self._get_axis_value(self.couch_keyword, "PatientSupportAngle")

    def _get_axis_value(self, axis_str: str, axis_dcm_attr: str) -> float:
        """Retrieve the value of the axis. This will first look in the file name for the value.
        If not in the filename then it will look in the DICOM metadata. If the value can be found in neither
        then a value of 0 is assumed.

        Parameters
        ----------
        axis_str : str
            The string to look for in the filename.
        axis_dcm_attr : str
            The DICOM attribute that should contain the axis value.

        Returns
        -------
        float
        """
        axis_found = False
        if self._use_filenames:
            filename = osp.basename(self.path)
            # see if the keyword is in the filename
            keyword_in_filename = axis_str.lower() in filename.lower()
            # if it's not there, then assume it's zero
            if not keyword_in_filename:
                axis = 0
                axis_found = True
            # if it is, then make sure it follows the naming convention of <axis###>
            else:
                match = re.search(
                    r"(?<={})\d+".format(axis_str.lower()), filename.lower()
                )
                if match is None:
                    raise ValueError(
                        f"The filename contains '{axis_str}' but could not read a number following it. Use the format '...{axis_str}<#>...'"
                    )
                else:
                    axis = float(match.group())
                    axis_found = True
        # try to interpret from DICOM data
        if not axis_found:
            try:
                axis = float(getattr(self.metadata, axis_dcm_attr))
            except AttributeError:
                axis = 0
        # if the value is close to 0 or 360 then peg at 0
        if is_close(axis, [0, 360], delta=1):
            return 0
        else:
            return axis


class FileImage(BaseImage):
    """An image from a "regular" file (.tif, .jpg, .bmp).

    Attributes
    ----------
    info : dict
        The info dictionary as generated by Pillow.
    sid : float
        The SID value as passed in upon construction.
    """

    def __init__(
        self,
        path: Union[str, Path, BinaryIO],
        *,
        dpi: Optional[float] = None,
        sid: Optional[float] = None,
        dtype: Optional[np.dtype] = None,
    ):
        """
        Parameters
        ----------
        path : str, file-object
            The path to the file or a data stream.
        dpi : int, float
            The dots-per-inch of the image, defined at isocenter.

            .. note:: If a DPI tag is found in the image, that value will override the parameter, otherwise this one
                will be used.
        sid : int, float
            The Source-to-Image distance in mm.
        dtype : numpy.dtype
            The data type to cast the array as.
        """
        super().__init__(path)
        pil_image = pImage.open(path)
        # convert to gray if need be
        if pil_image.mode not in ("F", "L", "1"):
            pil_image = pil_image.convert("F")
        self.info = pil_image.info
        if dtype is not None:
            self.array = np.array(pil_image, dtype=dtype)
        else:
            self.array = np.array(pil_image)
        self._dpi = dpi
        self.sid = sid

    @property
    def dpi(self) -> float:
        """The dots-per-inch of the image, defined at isocenter."""
        dpi = None
        for key in ("dpi", "resolution"):
            dpi = self.info.get(key)
            if dpi is not None:
                dpi = float(dpi[0])
                break
        if dpi is None:
            dpi = self._dpi
        if self.sid is not None and dpi is not None:
            dpi *= self.sid / 1000
        return dpi

    @property
    def dpmm(self) -> Optional[float]:
        """The Dots-per-mm of the image, defined at isocenter. E.g. if an EPID image is taken at 150cm SID,
        the dpmm will scale back to 100cm."""
        try:
            return self.dpi / MM_PER_INCH
        except TypeError:
            return


class ArrayImage(BaseImage):
    """An image constructed solely from a numpy array."""

    def __init__(
        self,
        array: np.array,
        *,
        dpi: float = None,
        sid: float = None,
        dtype=None,
    ):
        """
        Parameters
        ----------
        array : numpy.ndarray
            The image array.
        dpi : int, float
            The dots-per-inch of the image, defined at isocenter.

            .. note:: If a DPI tag is found in the image, that value will override the parameter, otherwise this one
                will be used.
        sid : int, float
            The Source-to-Image distance in mm.
        dtype : dtype, None, optional
            The data type to cast the image data as. If None, will use whatever raw image format is.
        """
        if dtype is not None:
            self.array = np.array(array, dtype=dtype)
        else:
            self.array = array
        self._dpi = dpi
        self.sid = sid

    @property
    def dpmm(self) -> Optional[float]:
        """The Dots-per-mm of the image, defined at isocenter. E.g. if an EPID image is taken at 150cm SID,
        the dpmm will scale back to 100cm."""
        try:
            return self.dpi / MM_PER_INCH
        except:
            return

    @property
    def dpi(self) -> Optional[float]:
        """The dots-per-inch of the image, defined at isocenter."""
        dpi = None
        if self._dpi is not None:
            dpi = self._dpi
            if self.sid is not None:
                dpi *= self.sid / 1000
        return dpi

    def __sub__(self, other):
        return ArrayImage(self.array - other.array)


class DicomImageStack:
    """A class that loads and holds a stack of DICOM images (e.g. a CT dataset). The class can take
    a folder or zip file and will read CT images. The images must all be the same size. Supports
    indexing to individual images.

    Attributes
    ----------
    images : list
        Holds instances of :class:`~pylinac.core.image.DicomImage`. Can be accessed via index;
        i.e. self[0] == self.images[0].

    Examples
    --------
    Load a folder of Dicom images
    >>> from pylinac import image
    >>> img_folder = r"folder/qa/cbct/june"
    >>> dcm_stack = image.DicomImageStack(img_folder)  # loads and sorts the images
    >>> dcm_stack.plot(3)  # plot the 3rd image

    Load a zip archive
    >>> img_folder_zip = r"archive/qa/cbct/june.zip"  # save space and zip your CBCTs
    >>> dcm_stack = image.DicomImageStack.from_zip(img_folder_zip)

    Load as a certain data type
    >>> dcm_stack_uint32 = image.DicomImageStack(img_folder, dtype=np.uint32)
    """

    images: List[ImageLike]

    def __init__(
        self,
        folder: Union[str, Path],
        dtype: Optional[np.dtype] = None,
        min_number: int = 39,
        check_uid: bool = True,
    ):
        """Load a folder with DICOM CT images.

        Parameters
        ----------
        folder : str
            Path to the folder.
        dtype : dtype, None, optional
            The data type to cast the image data as. If None, will use whatever raw image format is.
        """
        self.images = []
        paths = []
        # load in images in their received order
        if isinstance(folder, (list, tuple)):
            paths = folder
        elif osp.isdir(folder):
            for pdir, sdir, files in os.walk(folder):
                for file in files:
                    paths.append(osp.join(pdir, file))
        for path in paths:
            if self.is_image_slice(path):
                img = DicomImage(path, dtype=dtype)
                self.images.append(img)

        # check that at least 1 image was loaded
        if len(self.images) < 1:
            raise FileNotFoundError(
                f"No files were found in the specified location: {folder}"
            )

        # error checking
        if check_uid:
            self.images = self._check_number_and_get_common_uid_imgs(min_number)
        # sort according to physical order
        self.images.sort(key=lambda x: x.metadata.ImagePositionPatient[-1])

    @classmethod
    def from_zip(cls, zip_path: Union[str, Path], dtype: Optional[np.dtype] = None):
        """Load a DICOM ZIP archive.

        Parameters
        ----------
        zip_path : str
            Path to the ZIP archive.
        dtype : dtype, None, optional
            The data type to cast the image data as. If None, will use whatever raw image format is.
        """
        with TemporaryZipDirectory(zip_path) as tmpzip:
            obj = cls(tmpzip, dtype)
        return obj

    @staticmethod
    def is_image_slice(file: Union[str, Path]) -> bool:
        """Test if the file is a CT Image storage DICOM file."""
        try:
            ds = pydicom.dcmread(file, force=True, stop_before_pixels=True)
            return "Image Storage" in ds.SOPClassUID.name
        except (InvalidDicomError, AttributeError, MemoryError):
            return False

    def _check_number_and_get_common_uid_imgs(self, min_number: int) -> List:
        """Check that all the images are from the same study."""
        most_common_uid = Counter(
            i.metadata.SeriesInstanceUID for i in self.images
        ).most_common(1)[0]
        if most_common_uid[1] < min_number:
            raise ValueError(
                "The minimum number images from the same study were not found"
            )
        return [
            i for i in self.images if i.metadata.SeriesInstanceUID == most_common_uid[0]
        ]

    def plot(self, slice: int = 0) -> None:
        """Plot a slice of the DICOM dataset.

        Parameters
        ----------
        slice : int
            The slice to plot.
        """
        self.images[slice].plot()

    def roll(self, direction: str, amount: int):
        for img in self.images:
            img.roll(direction, amount)

    @property
    def metadata(self) -> pydicom.FileDataset:
        """The metadata of the first image; shortcut attribute. Only attributes that are common throughout the stack should be used,
        otherwise the individual image metadata should be used."""
        return self.images[0].metadata

    def __getitem__(self, item) -> DicomImage:
        return self.images[item]

    def __setitem__(self, key, value: DicomImage):
        self.images[key] = value

    def __len__(self):
        return len(self.images)


def gamma_2d(
    reference: np.ndarray,
    evaluation: np.ndarray,
    dose_to_agreement: float = 1,
    distance_to_agreement: int = 1,
    gamma_cap_value: float = 2,
    global_dose: bool = True,
    dose_threshold: float = 5,
    fill_value: float = np.nan,
) -> np.ndarray:
    """Compute a 2D gamma of two 2D numpy arrays. This does NOT do size or spatial resolution checking.
    It performs an element-by-element evaluation. It is the responsibility
    of the caller to ensure the reference and evaluation have comparable spatial resolution.

    The algorithm follows Table I of D. Low's 2004 paper: Evaluation of the gamma dose distribution comparison method: https://aapm.onlinelibrary.wiley.com/doi/epdf/10.1118/1.1598711

    This is similar to the gamma_1d function for profiles, except we must search a 2D grid around the reference point.

    Parameters
    ----------
    reference
        The reference 2D array.
    evaluation
        The evaluation 2D array.
    dose_to_agreement
        The dose to agreement in %. E.g. 1 is 1% of global reference max dose.
    distance_to_agreement
        The distance to agreement in **elements**. E.g. if the value is 4 this means 4 elements from the reference point under calculation.
        Must be >0
    gamma_cap_value
        The value to cap the gamma at. E.g. a gamma of 5.3 will get capped to 2. Useful for displaying data with a consistent range.
    global_dose
        Whether to evaluate the dose to agreement threshold based on the global max or the dose point under evaluation.
    dose_threshold
        The dose threshold as a number between 0 and 100 of the % of max dose under which a gamma is not calculated.
        This is not affected by the global/local dose normalization and the threshold value is evaluated against the global max dose, period.
    fill_value
        The value to give pixels that were not calculated because they were under the dose threshold. Default
        is NaN, but another option would be 0. If NaN, allows the user to calculate mean/median gamma over just the
        evaluated portion and not be skewed by 0's that should not be considered.
    """
    if reference.ndim != 2 or evaluation.ndim != 2:
        raise ValueError(
            f"Reference and evaluation arrays must be 2D. Got reference: {reference.ndim} and evaluation: {evaluation.ndim}"
        )
    threshold = reference.max() / 100 * dose_threshold
    # convert dose to agreement to % of global max; ignored later if local dose
    dose_ta = dose_to_agreement / 100 * reference.max()
    # pad eval array on both edges so our search does not go out of bounds
    eval_padded = np.pad(evaluation, distance_to_agreement, mode="edge")
    # iterate over each reference element, computing distance value and dose value
    gamma = np.zeros(reference.shape)
    for row_idx, row in enumerate(reference):
        for col_idx, ref_point in enumerate(row):
            # skip if below dose threshold
            if ref_point < threshold:
                gamma[row_idx, col_idx] = fill_value
                continue
            # use scikit-image to compute the indices of a disk around the reference point
            # we can then compute gamma over the eval points at these indices
            # unlike the 1D computation, we have to search at an index offset by the distance to agreement
            # we use DTA+1 in disk because it looks like the results are exclusive of edges.
            # https://scikit-image.org/docs/stable/api/skimage.draw.html#disk
            rs, cs = disk(
                (row_idx + distance_to_agreement, col_idx + distance_to_agreement),
                distance_to_agreement + 1,
            )

            capital_gammas = []
            for r, c in zip(rs, cs):
                eval_point = eval_padded[r, c]
                # for the distance, we compare the ref row/col to the eval padded matrix
                # but remember the padded array is padded by DTA, so to compare distances, we
                # have to cancel the offset we used for dose purposes.
                dist = math.dist(
                    (row_idx, col_idx),
                    (r - distance_to_agreement, c - distance_to_agreement),
                )
                dose = eval_point - ref_point
                if not global_dose:
                    dose_ta = dose_to_agreement / 100 * ref_point
                capital_gamma = math.sqrt(
                    dist ** 2 / distance_to_agreement ** 2 + dose ** 2 / dose_ta ** 2
                )
                capital_gammas.append(capital_gamma)
            gamma[row_idx, col_idx] = min(np.nanmin(capital_gammas), gamma_cap_value)
    return np.asarray(gamma)



# -*- coding: utf-8 -*-
"""
The Starshot module analyses a starshot image made of radiation spokes, whether gantry, collimator, MLC or couch.
It is based on ideas from `Depuydt et al <http://iopscience.iop.org/0031-9155/57/10/2997>`_
and `Gonzalez et al <http://dx.doi.org/10.1118/1.1755491>`_.

Features:

* **Analyze scanned film images, single EPID images, or a set of EPID images** -
  Any image that you can load in can be analyzed, including 1 or a set of EPID DICOM images and
  films that have been digitally scanned.
* **Any image size** - Have machines with different EPIDs? Scanned your film at different resolutions? No problem.
* **Dose/OD can be inverted** - Whether your device/image views dose as an increase in value or a decrease, pylinac
  will detect it and invert if necessary.
* **Automatic noise detection & correction** - Sometimes there's dirt on the scanned film; sometimes there's a dead pixel on the EPID.
  Pylinac will detect these spurious noise signals and can avoid or account for them.
* **Accurate, FWHM star line detection** - Pylinac uses not simply the maximum value to find the center of a star line,
  but analyzes the entire star profile to determine the center of the FWHM, ensuring small noise or maximum value bias is avoided.
* **Adaptive searching** - If you passed pylinac a set of parameters and a good result wasn't found, pylinac can recover and
  do an adaptive search by adjusting parameters to find a "reasonable" wobble.
"""
import copy
import dataclasses
import io
import webbrowser
from dataclasses import dataclass
from typing import Union, List, Optional, Tuple, BinaryIO

import argue
import matplotlib.pyplot as plt
import numpy as np
from scipy import optimize

from .core import image, pdf
from .core.geometry import Point, Line, Circle
from .core.io import get_url, TemporaryZipDirectory, retrieve_demo_file
from .core.profile import SingleProfile, CollapsedCircleProfile, Interpolation
from .core.utilities import ResultBase
from .settings import get_dicom_cmap


@dataclass
class StarshotResults(ResultBase):
    """This class should not be called directly. It is returned by the ``results_data()`` method.
    It is a dataclass under the hood and thus comes with all the dunder magic.

    Use the following attributes as normal class attributes.
    """

    tolerance_mm: float  #:
    circle_diameter_mm: float  #:
    circle_radius_mm: float  #:
    passed: bool  #:
    circle_center_x_y: Tuple[float, float]  #:


class Starshot:
    """Class that can determine the wobble in a "starshot" image, be it gantry, collimator,
    couch or MLC. The image can be a scanned film (TIF, JPG, etc) or a sequence of EPID DICOM images.

    Attributes
    ----------
    image : :class:`~pylinac.core.image.Image`
    circle_profile : :class:`~pylinac.starshot.StarProfile`
    lines : :class:`~pylinac.starshot.LineManager`
    wobble : :class:`~pylinac.starshot.Wobble`
    tolerance : :class:`~pylinac.starshot.Tolerance`

    Examples
    --------
    Run the demo:
        >>> Starshot.run_demo()

    Typical session:
        >>> img_path = r"C:/QA/Starshots/Coll.jpeg"
        >>> mystar = Starshot(img_path, dpi=105, sid=1000)
        >>> mystar.analyze()
        >>> print(mystar.results())
        >>> mystar.plot_analyzed_image()
    """

    def __init__(self, filepath: Union[str, BinaryIO], **kwargs):
        """
        Parameters
        ----------
        filepath
            The path to the image file.
        kwargs
            Passed to :func:`~pylinac.core.image.load`.
        """
        self.image = image.load(filepath, **kwargs)
        self.wobble = Wobble()
        self.tolerance = 1
        if self.image.dpmm is None:
            raise ValueError(
                "DPI was not a tag in the image nor was it passed in. Please pass a DPI value"
            )
        if self.image.sid is None:
            raise ValueError(
                "Source-to-Image distance was not an image tag and was not passed in. Please pass an SID value."
            )

    @classmethod
    def from_url(cls, url: str, **kwargs):
        """Instantiate from a URL.

        Parameters
        ----------
        url : str
            URL of the raw file.
        kwargs
            Passed to :func:`~pylinac.core.image.load`.
        """
        filename = get_url(url)
        return cls(filename, **kwargs)

    @classmethod
    def from_demo_image(cls):
        """Construct a Starshot instance and load the demo image."""
        demo_file = retrieve_demo_file(name="starshot.tif")
        return cls(demo_file, sid=1000)

    @classmethod
    def from_multiple_images(cls, filepath_list: list, **kwargs):
        """Construct a Starshot instance and load in and combine multiple images.

        Parameters
        ----------
        filepath_list : iterable
            An iterable of file paths to starshot images that are to be superimposed.
        kwargs
            Passed to :func:`~pylinac.core.image.load_multiples`.
        """
        obj = cls.from_demo_image()
        obj.image = image.load_multiples(filepath_list, **kwargs)
        return obj

    @classmethod
    def from_zip(cls, zip_file: str, **kwargs):
        """Construct a Starshot instance from a ZIP archive.

        Parameters
        ----------
        zip_file : str
            Points to the ZIP archive. Can contain a single or multiple images. If multiple images
            the images are combined and thus should be from the same test sequence.
        kwargs
            Passed to :func:`~pylinac.core.image.load_multiples`.
        """
        with TemporaryZipDirectory(zip_file) as tmpdir:
            image_files = image.retrieve_image_files(tmpdir)
            if not image_files:
                raise IndexError(f"No valid starshot images were found in {zip_file}")
            if len(image_files) > 1:
                return cls.from_multiple_images(image_files, **kwargs)
            else:
                return cls(image_files[0], **kwargs)

    def _get_reasonable_start_point(self) -> Point:
        """Set the algorithm starting point automatically.

        Notes
        -----
        The determination of an automatic start point is accomplished by finding the Full-Width-80%-Max.
        Finding the maximum pixel does not consistently work, esp. in the presence of a pin prick. The
        FW80M is a more consistent metric for finding a good start point.
        """
        # sum the image along each axis within the central 1/3 (avoids outlier influence from say, gantry shots)
        top_third = int(self.image.array.shape[0] / 3)
        bottom_third = int(top_third * 2)
        left_third = int(self.image.array.shape[1] / 3)
        right_third = int(left_third * 2)
        central_array = self.image.array[top_third:bottom_third, left_third:right_third]

        x_sum = np.sum(central_array, 0)
        y_sum = np.sum(central_array, 1)

        # Calculate Full-Width, 80% Maximum center
        fwxm_x_point = (
            SingleProfile(x_sum, interpolation=Interpolation.NONE).fwxm_data(80)[
                "center index (rounded)"
            ]
            + left_third
        )
        fwxm_y_point = (
            SingleProfile(y_sum, interpolation=Interpolation.NONE).fwxm_data(80)[
                "center index (rounded)"
            ]
            + top_third
        )
        center_point = Point(fwxm_x_point, fwxm_y_point)
        return center_point

    @argue.bounds(radius=(0.2, 0.95), min_peak_height=(0.05, 0.95))
    def analyze(
        self,
        radius: float = 0.85,
        min_peak_height: float = 0.25,
        tolerance: float = 1.0,
        start_point: Optional[Union[Point, tuple]] = None,
        fwhm: bool = True,
        recursive: bool = True,
        invert: bool = False,
    ):
        """Analyze the starshot image.

        Analyze finds the minimum radius and center of a circle that touches all the lines
        (i.e. the wobble circle diameter and wobble center).

        Parameters
        ----------
        radius : float, optional
            Distance in % between starting point and closest image edge; used to build the circular profile which finds
            the radiation lines. Must be between 0.05 and 0.95.
        min_peak_height : float, optional
            The percentage minimum height a peak must be to be considered a valid peak. A lower value catches
            radiation peaks that vary in magnitude (e.g. different MU delivered or gantry shot), but could also pick up noise.
            If necessary, lower value for gantry shots and increase for noisy images.
        tolerance : int, float, optional
            The tolerance in mm to test against for a pass/fail result.
        start_point : 2-element iterable, optional
            The point where the algorithm should center the circle profile, given as (x-value, y-value).
            If None (default), will search for a reasonable maximum point nearest the center of the image.
        fwhm : bool
            If True (default), the center of the FWHM of the spokes will be determined.
            If False, the peak value location is used as the spoke center.

            .. note:: In practice, this ends up being a very small difference. Set to false if peak locations are offset or unexpected.
        recursive : bool
            If True (default), will recursively search for a "reasonable" wobble, meaning the wobble radius is
            <3mm. If the wobble found was unreasonable,
            the minimum peak height is iteratively adjusted from low to high at the passed radius.
            If for all peak heights at the given radius the wobble is still unreasonable, the
            radius is then iterated over from most distant inward, iterating over minimum peak heights at each radius.
            If False, will simply return the first determined value or raise error if a reasonable wobble could not be determined.

            .. warning:: It is strongly recommended to leave this setting at True.

        invert : bool
            Whether to force invert the image values. This should be set to True if the automatically-determined
            pylinac inversion is incorrect.

        Raises
        ------
        RuntimeError
            If a reasonable wobble value was not found.
        """
        self.tolerance = tolerance
        self.image.check_inversion_by_histogram(percentiles=[4, 50, 96])
        if invert:
            self.image.invert()

        if start_point is None:
            start_point = self._get_reasonable_start_point()

        self._get_reasonable_wobble(
            start_point, fwhm, min_peak_height, radius, recursive
        )

    def _get_reasonable_wobble(
        self, start_point, fwhm, min_peak_height, radius, recursive
    ):
        """Determine a wobble that is "reasonable". If recursive is false, the first iteration will be passed,
        otherwise the parameters will be tweaked to search for a reasonable wobble."""
        wobble_unreasonable = True
        focus_point = copy.copy(start_point)
        peak_gen = get_peak_height()
        radius_gen = get_radius()
        while wobble_unreasonable:
            try:
                self.circle_profile = StarProfile(
                    self.image, focus_point, radius, min_peak_height, fwhm
                )
                if (len(self.circle_profile.peaks) < 6) or (
                    len(self.circle_profile.peaks) % 2 != 0
                ):
                    raise ValueError
                self.lines = LineManager(self.circle_profile.peaks)
                self._find_wobble_minimize()
            except ValueError:
                if not recursive:
                    raise RuntimeError(
                        "The algorithm was unable to properly detect the radiation lines. Try setting "
                        "recursive to True or lower the minimum peak height"
                    )
                else:
                    try:
                        min_peak_height = next(peak_gen)
                    except StopIteration:
                        # if no height setting works, change the radius and reset the height
                        try:
                            radius = next(radius_gen)
                            peak_gen = get_peak_height()
                        except StopIteration:
                            raise RuntimeError(
                                "The algorithm was unable to determine a reasonable wobble. Try setting "
                                "recursive to False and manually adjusting algorithm parameters"
                            )

            else:  # if no errors are raised
                # set the focus point to the wobble minimum
                # focus_point = self.wobble.center
                # finally:
                # stop after first iteration if not recursive
                if not recursive:
                    wobble_unreasonable = False
                # otherwise, check if the wobble is reasonable
                else:
                    # if so, stop
                    if self.wobble.diameter_mm < 2:
                        focus_near_center = (
                            self.wobble.center.distance_to(focus_point) < 5
                        )
                        if focus_near_center:
                            wobble_unreasonable = False
                        else:
                            focus_point = self.wobble.center
                    # otherwise, iterate through peak height
                    else:
                        try:
                            min_peak_height = next(peak_gen)
                        except StopIteration:
                            # if no height setting works, change the radius and reset the height
                            try:
                                radius = next(radius_gen)
                                peak_gen = get_peak_height()
                            except StopIteration:
                                raise RuntimeError(
                                    "The algorithm was unable to determine a reasonable wobble. Try setting "
                                    "recursive to False and manually adjusting algorithm parameters"
                                )

    def _find_wobble_minimize(self) -> None:
        """Find the minimum distance wobble location and radius to all radiation lines.

        The minimum is found using a scipy minimization function.
        """
        # starting point
        sp = copy.copy(self.circle_profile.center)

        def distance(p, lines):
            """Calculate the maximum distance to any line from the given point."""
            return max(line.distance_to(Point(p[0], p[1])) for line in lines)

        res = optimize.minimize(
            distance,
            sp.as_array(),
            args=(self.lines,),
            method="Nelder-Mead",
            options={"fatol": 0.001},
        )
        # res = optimize.least_squares(distance, sp.as_array(), args=(self.lines,), ftol=0.001)

        self.wobble.radius = res.fun
        self.wobble.radius_mm = res.fun / self.image.dpmm
        self.wobble.center = Point(res.x[0], res.x[1])

    @property
    def passed(self) -> bool:
        """Boolean specifying whether the determined wobble was within tolerance."""
        return self.wobble.radius_mm * 2 < self.tolerance

    @property
    def _passfail_str(self) -> str:
        """Return a pass/fail string."""
        return "PASS" if self.passed else "FAIL"

    def results(self) -> str:
        """Return the results of the analysis.

        Returns
        -------
        string
            A string with a statement of the minimum circle.
        """
        string = (
            f"\nResult: {self._passfail_str} \n\n"
            + f"The minimum circle that touches all the star lines has a diameter of {self.wobble.radius_mm*2:2.3f} mm. \n\n"
            + f"The center of the minimum circle is at {self.wobble.center.x:3.1f}, {self.wobble.center.y:3.1f}"
        )
        return string

    def results_data(self, as_dict: bool = False) -> Union[StarshotResults, dict]:
        """Present the results data and metadata as a dataclass or dict.
        The default return type is a dataclass."""
        data = StarshotResults(
            tolerance_mm=self.tolerance,
            circle_diameter_mm=self.wobble.radius_mm * 2,
            circle_radius_mm=self.wobble.radius_mm,
            circle_center_x_y=(self.wobble.center.x, self.wobble.center.y),
            passed=self.passed,
        )
        if as_dict:
            return dataclasses.asdict(data)
        return data

    def plot_analyzed_image(self, show: bool = True, **plt_kwargs: dict):
        """Draw the star lines, profile circle, and wobble circle on a matplotlib figure.

        Parameters
        ----------
        show : bool
            Whether to actually show the image.
        plt_kwargs : dict
            Keyword args passed to the plt.subplots() method. Allows one to set things like figure size.
        """
        fig, axes = plt.subplots(ncols=2, **plt_kwargs)
        subimages = ("whole", "wobble")
        titles = ("Analyzed Image", "Wobble Circle")

        # show images
        for ax, subimage, title in zip(axes, subimages, titles):
            self.plot_analyzed_subimage(ax=ax, show=False, subimage=subimage)
            ax.set_title(title)

        if show:
            plt.show()

    def plot_analyzed_subimage(
        self,
        subimage: str = "wobble",
        ax: Optional[plt.Axes] = None,
        show: bool = True,
        **plt_kwargs: dict,
    ):
        """Plot a subimage of the starshot analysis. Current options are the zoomed out image and the zoomed in image.

        Parameters
        ----------
        subimage : str
            If 'wobble', will show a zoomed in plot of the wobble circle.
            Any other string will show the zoomed out plot.
        ax : None, matplotlib Axes
            If None (default), will create a new figure to plot on, otherwise plot to the passed axes.
        show : bool
            Whether to actually show the image.
        plt_kwargs : dict
            Keyword args passed to the plt.figure() method. Allows one to set things like figure size. Only used if ax is not passed.
        """
        if ax is None:
            fig, ax = plt.subplots(**plt_kwargs)
        # show analyzed image
        ax.imshow(self.image.array, cmap=get_dicom_cmap())
        self.lines.plot(ax)
        self.wobble.plot2axes(ax, edgecolor="green")
        self.circle_profile.plot2axes(ax, edgecolor="green")
        ax.autoscale(tight=True)
        ax.axis("off")

        # zoom in if wobble plot
        if subimage == "wobble":
            xlims = [
                self.wobble.center.x + self.wobble.diameter,
                self.wobble.center.x - self.wobble.diameter,
            ]
            ylims = [
                self.wobble.center.y + self.wobble.diameter,
                self.wobble.center.y - self.wobble.diameter,
            ]
            ax.set_xlim(xlims)
            ax.set_ylim(ylims)
            ax.axis("on")

        if show:
            plt.show()

    def save_analyzed_image(self, filename: str, **kwargs):
        """Save the analyzed image plot to a file.

        Parameters
        ----------
        filename : str, IO stream
            The filename to save as. Format is deduced from string extention, if there is one. E.g. 'mystar.png' will
            produce a PNG image.
        kwargs
            All other kwargs are passed to plt.savefig().
        """
        self.plot_analyzed_image(show=False)
        plt.savefig(filename, **kwargs)

    def save_analyzed_subimage(self, filename: str, subimage: str = "wobble", **kwargs):
        """Save the analyzed subimage to a file.

        Parameters
        ----------
        filename : str, file-object
            Where to save the file to.
        subimage : str
            If 'wobble', will show a zoomed in plot of the wobble circle.
            Any other string will show the zoomed out plot.
        kwargs
            Passed to matplotlib.
        """
        self.plot_analyzed_subimage(subimage=subimage, show=False)
        plt.savefig(filename, **kwargs)

    def publish_pdf(
        self,
        filename: Union[str, BinaryIO],
        notes: Optional[Union[str, List[str]]] = None,
        open_file: bool = False,
        metadata: Optional[dict] = None,
    ):
        """Publish (print) a PDF containing the analysis, images, and quantitative results.

        Parameters
        ----------
        filename : (str, file-like object}
            The file to write the results to.
        notes : str, list of strings
            Text; if str, prints single line.
            If list of strings, each list item is printed on its own line.
        open_file : bool
            Whether to open the file using the default program after creation.
        metadata : dict
            Extra data to be passed and shown in the PDF. The key and value will be shown with a colon.
            E.g. passing {'Author': 'James', 'Unit': 'TrueBeam'} would result in text in the PDF like:
            --------------
            Author: James
            Unit: TrueBeam
            --------------
        """
        canvas = pdf.PylinacCanvas(
            filename, page_title="Starshot Analysis", metadata=metadata
        )
        for img, height in zip(("wobble", "asdf"), (2, 11.5)):
            data = io.BytesIO()
            self.save_analyzed_subimage(data, img)
            canvas.add_image(data, location=(4, height), dimensions=(13, 13))
        text = [
            "Starshot results:",
            f"Source-to-Image Distance (mm): {self.image.sid:2.0f}",
            f"Tolerance (mm): {self.tolerance:2.1f}",
            f"Minimum circle diameter (mm): {self.wobble.radius_mm*2:2.2f}",
        ]
        canvas.add_text(text=text, location=(10, 25.5), font_size=12)
        if notes is not None:
            canvas.add_text(text="Notes:", location=(1, 5.5), font_size=14)
            canvas.add_text(text=notes, location=(1, 5))
        canvas.finish()

        if open_file:
            webbrowser.open(filename)

    @staticmethod
    def run_demo():
        """Demonstrate the Starshot module using the demo image."""
        star = Starshot.from_demo_image()
        star.analyze()
        print(star.results())
        star.plot_analyzed_image()


class Wobble(Circle):
    """A class that holds the wobble information of the Starshot analysis.

    Attributes
    ----------
    radius_mm : The radius of the Circle in **mm**.
    """

    def __init__(self, center_point=None, radius=None):
        super().__init__(center_point=center_point, radius=radius)
        self.radius_mm = 0  # The radius of the wobble in mm; as opposed to pixels.

    @property
    def diameter_mm(self) -> float:
        """Diameter of the wobble in mm."""
        return self.radius_mm * 2


class LineManager:
    """Manages the radiation lines found."""

    def __init__(self, points: List[Point]):
        """
        Parameters
        ----------
        points :
            The peak points found by the StarProfile
        """
        self.lines = []
        self.construct_rad_lines(points)

    def __getitem__(self, item):
        return self.lines[item]

    def __len__(self):
        return len(self.lines)

    def construct_rad_lines(self, points: List[Point]):
        """Find and match the positions of peaks in the circle profile (radiation lines)
            and map their positions to the starshot image.

        Radiation lines are found by finding the FWHM of the radiation spokes, then matching them
        to form lines.

        Returns
        -------
        lines : list
            A list of Lines (radiation lines) found.

        See Also
        --------
        Starshot.analyze() : min_peak_height parameter info
        core.profile.CircleProfile.find_FWXM_peaks : min_peak_distance parameter info.
        geometry.Line : returning object
        """
        self.match_points(points)

    def match_points(self, points: List[Point]):
        """Match the peaks found to the same radiation lines.

        Peaks are matched by connecting the existing peaks based on an offset of peaks. E.g. if there are
        12 peaks, there must be 6 radiation lines. Furthermore, assuming star lines go all the way across the CAX,
        the 7th peak will be the opposite peak of the 1st peak, forming a line. This method is robust to
        starting points far away from the real center.
        """
        num_rad_lines = int(len(points) / 2)
        offset = num_rad_lines
        self.lines = [
            Line(points[line], points[line + offset]) for line in range(num_rad_lines)
        ]

    def plot(self, axis: plt.Axes):
        """Plot the lines to the axis."""
        for line in self.lines:
            line.plot2axes(axis, color="blue")


class StarProfile(CollapsedCircleProfile):
    """Class that holds and analyzes the circular profile which finds the radiation lines."""

    def __init__(self, image, start_point, radius, min_peak_height, fwhm):
        radius = self._convert_radius_perc2pix(image, start_point, radius)
        super().__init__(
            center=start_point,
            radius=radius,
            image_array=image.array,
            width_ratio=0.1,
            sampling_ratio=3,
        )
        self.get_peaks(min_peak_height, fwhm=fwhm)

    @staticmethod
    def _convert_radius_perc2pix(image, start_point, radius):
        """Convert a percent radius to distance in pixels, based on the distance from center point to image
            edge.

        Parameters
        ----------
        radius : float
            The radius ratio (e.g. 0.5).
        """
        return image.dist2edge_min(start_point) * radius

    def _roll_prof_to_midvalley(self) -> int:
        """Roll the circle profile so that its edges are not near a radiation line.
        This is a prerequisite for properly finding star lines.
        """
        roll_amount = np.where(self.values == self.values.min())[0][0]
        self.roll(roll_amount)
        return roll_amount

    def get_peaks(self, min_peak_height, min_peak_distance=0.02, fwhm=True):
        """Determine the peaks of the profile."""
        self._roll_prof_to_midvalley()
        self.filter(size=0.003, kind="gaussian")
        self.ground()
        if fwhm:
            self.find_fwxm_peaks(
                threshold=min_peak_height, min_distance=min_peak_distance
            )
        else:
            self.find_peaks(min_peak_height, min_peak_distance)


def get_peak_height():
    for height in np.linspace(0.05, 0.95, 10):
        yield height


def get_radius():
    for radius in np.linspace(0.95, 0.1, 10):
        yield radius
